"""
ASCII art generation for Terminaide using the ansi-shadow font.

This module provides the terminascii function for generating ASCII art banners,
consolidating the original utils.py and vendor/bigfont_slim.py into a single file.
"""

from __future__ import print_function
from __future__ import division

import functools
import os
import re
import logging
import operator
import inspect
import sys
import copy
from typing import Optional

try:  # py 2/3 compatible import
    from itertools import izip as zip
except ImportError:
    pass

logger = logging.getLogger("terminaide")


# === Base Classes ===

class BaseObject(object):
    """Ancestor class for all other classes."""

    def __init__(self, *args, **kwargs):
        super(BaseObject, self).__init__()

        if args or kwargs:
            logging.debug("%s has unused parameters %s, %s"
                          % (self, args, kwargs))


# === Decorators ===

def trace(func):
    """Decorator to print arguments to called function and the return value."""

    @functools.wraps(func)
    def decorator(*args, **kwargs):
        logging.debug("%s(%s, %s)" % (func.__name__, args, kwargs))
        ret = func(*args, **kwargs)
        logging.debug("%s returned %s" % (func.__name__, ret))
        return ret

    return decorator


# === Smooshing ===

def _smoosh_spaces(left, right):
    if right == ' ':
        return left
    return right


class Smoosher(BaseObject):
    """Uses various rules to combine characters into a single character."""

    def __init__(self, **kwargs):
        super(Smoosher, self).__init__(**kwargs)
        # Only space smooshing is used in the current codebase
        self.rules = [_smoosh_spaces]

    def smoosh(self, left, right):
        """Smoosh single characters according to smooshing rules."""

        outchars = []
        for lc, rc in zip(left, right):
            for rule in self.rules:
                sc = rule(lc, rc)
                if sc is not None:
                    outchars.append(sc)
                    break

        return ''.join(outchars)


# === BigLetter ===

class BigLetter(BaseObject):
    """
    Represents a single letter in a font.
    """

    def __init__(self, lines, hardblank='$', rules=None, **kwargs):
        super(BigLetter, self).__init__(**kwargs)
        self._set_lines(lines)
        self.hardblank = hardblank
        if rules is None:
            self.rules = Smoosher(hardblank=hardblank)
        else:
            self.rules = rules

    def _set_lines(self, lines):
        self.lines = list(lines)
        self.height = len(self.lines)
        self.maxwidth = max(self.lines, key=len)

    def __str__(self):
        out = "\n".join(self.lines)
        return re.sub(re.escape(self.hardblank), ' ', out)  # remove hardblanks

    def __add__(self, other):
        """Shortcut to kern()."""
        return self.kern(other)

    def __iter__(self):
        for line in self.lines:
            yield line

    @trace
    def horizontal_space(self, other):
        """Returns the smallest amount of horizontal space between
        this letter's right side and another letter."""
        minspace = None
        for lrow, rrow in zip(self, other):
            ls = lrow.rstrip()
            rs = rrow.lstrip()
            lstripped = len(lrow) - len(ls)
            rstripped = len(rrow) - len(rs)
            separation = lstripped + rstripped
            if minspace is None or separation < minspace:
                minspace = separation
        return minspace

    def kern(self, other):
        """Overlap two letters until they touch, and return a new letter."""
        overlap = self.horizontal_space(other)
        return self.push(other, overlap=overlap)

    def push(self, other, overlap=1):
        """Push two letters together into a new one."""
        newlines = []
        for s, o in zip(self.lines, other.lines):
            if overlap < 1:
                newlines.append(s + o)
            else:
                leftchars = s[:-overlap]
                rightchars = o[overlap:]
                leftoverlap = s[-overlap:]
                rightoverlap = o[:overlap]
                overlapped = self.rules.smoosh(leftoverlap, rightoverlap)
                newlines.append(leftchars + overlapped + rightchars)
        newletter = copy.copy(self)
        newletter._set_lines(newlines)
        return newletter


# === BigFont and Font Loading ===

class BigFontError(Exception):
    pass


class BigFont(BaseObject):
    """Stores all characters from a font as a list of BigLetters."""

    def __init__(self, data=None, name=None, nonprintable=32, eol="\n", **kwargs):
        super(BigFont, self).__init__(**kwargs)
        self.nonprintable = nonprintable
        try:
            self.letters = self._extract_letters(data)
        except:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            logging.warning("load failed, %s: %s " % (exc_type, exc_value))
            raise BigFontError("font data did not make sense")
        # self.renderfcn = kern
        self.eol = eol
        self.name = name
        self.smooshrules = None  # make rules from header line
        self.raise_missing = True

    def _extract_letters(self, data):
        """Extract list of BigLetters from flf file data"""
        if data is None:
            return None
        elif isinstance(data, str):
            lines = re.split("\r?\n", data)
        else:
            lines = [line.rstrip() for line in data]

        header = self._parse_header(lines[0])
        # last char in first data line is the endchar
        endchar = lines[header['comment_lines'] + 1][-1]
        buf = []
        maxchar = 255
        letters = [None] * (maxchar + 1)
        index = 32  # first required character index
        optional_chars = re.compile(r"""^(\d+)\b.+(?<!%s)$""" % endchar)

        for line in lines:
            if index > 126:  # required chars
                m = re.match(optional_chars, line)
                if m:
                    index = int(m.group(1))
                    continue

            if index > maxchar:
                continue

            if len(line) > 1 and line[-2:] == (endchar * 2):
                buf.append(line[:-2])
                letters[index] = BigLetter(buf, hardblank=header['hardblank'])
                logging.debug("loaded character %s:\n%s" % (index, letters[index]))
                index += 1
                buf = []
            elif len(line) > 0 and line[-1] == endchar:
                buf.append(line[:-1])

        # copy required german characters to their correct positions
        # from 127-133
        moveto = (196, 214, 220, 228, 246, 252, 223)
        for idx, char in enumerate(moveto):
            if letters[char] is None and letters[idx + 127] is not None:
                letters[char] = letters[idx + 127]
                letters[idx + 127] = None

        return letters

    @trace
    def _parse_header(self, hdr):
        """Retrieve info from FIGfont header line, return as dict."""
        out = {}
        if len(hdr) < 5 or hdr[0:5] != "flf2a":
            raise BigFontError("did not understand header %s" % hdr)
        fields = hdr.split()
        out['signature'] = fields[0][0:5]
        out['hardblank'] = fields[0][5]
        out['height'] = int(fields[1])
        out['baseline'] = int(fields[2])
        out['max_length'] = int(fields[3])
        out['old_layout'] = int(fields[4])
        out['comment_lines'] = int(fields[5])
        if len(fields) > 6:
            out['print_direction'] = int(fields[6])
            out['full_layout'] = int(fields[7])
            out['codetag_count'] = int(fields[8])

        return out

    def __getitem__(self, key):
        """Access a BigLetter from the character representation."""
        try:
            return self.letters[ord(key)]
        except IndexError:
            if self.raise_missing:
                raise KeyError("%s is not present in font" % key)
            else:
                return self.letters[ord('?')]  # fixme, return an empty Bigletter?

    def render(self, s):
        """Return string rendered in the font, suitable for printing."""
        return functools.reduce(operator.add, [self[c] for c in s])


# === Font Data (embedded ansi-shadow font) ===

ANSI_SHADOW_FONT_DATA = r"""flf2a$ 7 7 13 0 7 0 64 0
Font Author: ?

More Info:

https://web.archive.org/web/20120819044459/http://www.roysac.com/thedrawfonts-tdf.asp

FIGFont created with: http://patorjk.com/figfont-editor
$  $@
$  $@
$  $@
$  $@
$  $@
$  $@
$  $@@
██╗@
██║@
██║@
╚═╝@
██╗@
╚═╝@
   @@
@
@
@
@
@
@
@@
 ██╗ ██╗ @
████████╗@
╚██╔═██╔╝@
████████╗@
╚██╔═██╔╝@
 ╚═╝ ╚═╝ @
         @@
▄▄███▄▄·@
██╔════╝@
███████╗@
╚════██║@
███████║@
╚═▀▀▀══╝@
        @@
██╗ ██╗@
╚═╝██╔╝@
  ██╔╝ @
 ██╔╝  @
██╔╝██╗@
╚═╝ ╚═╝@
       @@
   ██╗   @
   ██║   @
████████╗@
██╔═██╔═╝@
██████║  @
╚═════╝  @
         @@
@
@
@
@
@
@
@@
 ██╗@
██╔╝@
██║ @
██║ @
╚██╗@
 ╚═╝@
    @@
██╗ @
╚██╗@
 ██║@
 ██║@
██╔╝@
╚═╝ @
    @@
      @
▄ ██╗▄@
 ████╗@
▀╚██╔▀@
  ╚═╝ @
      @
      @@
@
@
@
@
@
@
@@
   @
   @
   @
   @
▄█╗@
╚═╝@
   @@
      @
      @
█████╗@
╚════╝@
      @
      @
      @@
   @
   @
   @
   @
██╗@
╚═╝@
   @@
    ██╗@
   ██╔╝@
  ██╔╝ @
 ██╔╝  @
██╔╝   @
╚═╝    @
       @@
 ██████╗ @
██╔═████╗@
██║██╔██║@
████╔╝██║@
╚██████╔╝@
 ╚═════╝ @
         @@
 ██╗@
███║@
╚██║@
 ██║@
 ██║@
 ╚═╝@
    @@
██████╗ @
╚════██╗@
 █████╔╝@
██╔═══╝ @
███████╗@
╚══════╝@
        @@
██████╗ @
╚════██╗@
 █████╔╝@
 ╚═══██╗@
██████╔╝@
╚═════╝ @
        @@
██╗  ██╗@
██║  ██║@
███████║@
╚════██║@
     ██║@
     ╚═╝@
        @@
███████╗@
██╔════╝@
███████╗@
╚════██║@
███████║@
╚══════╝@
        @@
 ██████╗ @
██╔════╝ @
███████╗ @
██╔═══██╗@
╚██████╔╝@
 ╚═════╝ @
         @@
███████╗@
╚════██║@
    ██╔╝@
   ██╔╝ @
   ██║  @
   ╚═╝  @
        @@
 █████╗ @
██╔══██╗@
╚█████╔╝@
██╔══██╗@
╚█████╔╝@
 ╚════╝ @
        @@
 █████╗ @
██╔══██╗@
╚██████║@
 ╚═══██║@
 █████╔╝@
 ╚════╝ @
        @@
   @
██╗@
╚═╝@
██╗@
╚═╝@
   @
   @@
   @
██╗@
╚═╝@
▄█╗@
▀═╝@
   @
   @@
  ██╗@
 ██╔╝@
██╔╝ @
╚██╗ @
 ╚██╗@
  ╚═╝@
     @@
@
@
@
@
@
@
@@
██╗  @
╚██╗ @
 ╚██╗@
 ██╔╝@
██╔╝ @
╚═╝  @
     @@
██████╗ @
╚════██╗@
  ▄███╔╝@
  ▀▀══╝ @
  ██╗   @
  ╚═╝   @
        @@
 ██████╗ @
██╔═══██╗@
██║██╗██║@
██║██║██║@
╚█║████╔╝@
 ╚╝╚═══╝ @
         @@
 █████╗ @
██╔══██╗@
███████║@
██╔══██║@
██║  ██║@
╚═╝  ╚═╝@
        @@
██████╗ @
██╔══██╗@
██████╔╝@
██╔══██╗@
██████╔╝@
╚═════╝ @
        @@
 ██████╗@
██╔════╝@
██║     @
██║     @
╚██████╗@
 ╚═════╝@
        @@
██████╗ @
██╔══██╗@
██║  ██║@
██║  ██║@
██████╔╝@
╚═════╝ @
        @@
███████╗@
██╔════╝@
█████╗  @
██╔══╝  @
███████╗@
╚══════╝@
        @@
███████╗@
██╔════╝@
█████╗  @
██╔══╝  @
██║     @
╚═╝     @
        @@
 ██████╗ @
██╔════╝ @
██║  ███╗@
██║   ██║@
╚██████╔╝@
 ╚═════╝ @
         @@
██╗  ██╗@
██║  ██║@
███████║@
██╔══██║@
██║  ██║@
╚═╝  ╚═╝@
        @@
██╗@
██║@
██║@
██║@
██║@
╚═╝@
   @@
     ██╗@
     ██║@
     ██║@
██   ██║@
╚█████╔╝@
 ╚════╝ @
        @@
██╗  ██╗@
██║ ██╔╝@
█████╔╝ @
██╔═██╗ @
██║  ██╗@
╚═╝  ╚═╝@
        @@
██╗     @
██║     @
██║     @
██║     @
███████╗@
╚══════╝@
        @@
███╗   ███╗@
████╗ ████║@
██╔████╔██║@
██║╚██╔╝██║@
██║ ╚═╝ ██║@
╚═╝     ╚═╝@
           @@
███╗   ██╗@
████╗  ██║@
██╔██╗ ██║@
██║╚██╗██║@
██║ ╚████║@
╚═╝  ╚═══╝@
          @@
 ██████╗ @
██╔═══██╗@
██║   ██║@
██║   ██║@
╚██████╔╝@
 ╚═════╝ @
         @@
██████╗ @
██╔══██╗@
██████╔╝@
██╔═══╝ @
██║     @
╚═╝     @
        @@
 ██████╗ @
██╔═══██╗@
██║   ██║@
██║▄▄ ██║@
╚██████╔╝@
 ╚══▀▀═╝ @
         @@
██████╗ @
██╔══██╗@
██████╔╝@
██╔══██╗@
██║  ██║@
╚═╝  ╚═╝@
        @@
███████╗@
██╔════╝@
███████╗@
╚════██║@
███████║@
╚══════╝@
        @@
████████╗@
╚══██╔══╝@
   ██║   @
   ██║   @
   ██║   @
   ╚═╝   @
         @@
██╗   ██╗@
██║   ██║@
██║   ██║@
██║   ██║@
╚██████╔╝@
 ╚═════╝ @
         @@
██╗   ██╗@
██║   ██║@
██║   ██║@
╚██╗ ██╔╝@
 ╚████╔╝ @
  ╚═══╝  @
         @@
██╗    ██╗@
██║    ██║@
██║ █╗ ██║@
██║███╗██║@
╚███╔███╔╝@
 ╚══╝╚══╝ @
          @@
██╗  ██╗@
╚██╗██╔╝@
 ╚███╔╝ @
 ██╔██╗ @
██╔╝ ██╗@
╚═╝  ╚═╝@
        @@
██╗   ██╗@
╚██╗ ██╔╝@
 ╚████╔╝ @
  ╚██╔╝  @
   ██║   @
   ╚═╝   @
         @@
███████╗@
╚══███╔╝@
  ███╔╝ @
 ███╔╝  @
███████╗@
╚══════╝@
        @@
███╗@
██╔╝@
██║ @
██║ @
███╗@
╚══╝@
    @@
@
@
@
@
@
@
@@
███╗@
╚██║@
 ██║@
 ██║@
███║@
╚══╝@
    @@
 ███╗ @
██╔██╗@
╚═╝╚═╝@
      @
      @
      @
      @@
        @
        @
        @
        @
███████╗@
╚══════╝@
        @@
@
@
@
@
@
@
@@
 █████╗ @
██╔══██╗@
███████║@
██╔══██║@
██║  ██║@
╚═╝  ╚═╝@
        @@
██████╗ @
██╔══██╗@
██████╔╝@
██╔══██╗@
██████╔╝@
╚═════╝ @
        @@
 ██████╗@
██╔════╝@
██║     @
██║     @
╚██████╗@
 ╚═════╝@
        @@
██████╗ @
██╔══██╗@
██║  ██║@
██║  ██║@
██████╔╝@
╚═════╝ @
        @@
███████╗@
██╔════╝@
█████╗  @
██╔══╝  @
███████╗@
╚══════╝@
        @@
███████╗@
██╔════╝@
█████╗  @
██╔══╝  @
██║     @
╚═╝     @
        @@
 ██████╗ @
██╔════╝ @
██║  ███╗@
██║   ██║@
╚██████╔╝@
 ╚═════╝ @
         @@
██╗  ██╗@
██║  ██║@
███████║@
██╔══██║@
██║  ██║@
╚═╝  ╚═╝@
        @@
██╗@
██║@
██║@
██║@
██║@
╚═╝@
   @@
     ██╗@
     ██║@
     ██║@
██   ██║@
╚█████╔╝@
 ╚════╝ @
        @@
██╗  ██╗@
██║ ██╔╝@
█████╔╝ @
██╔═██╗ @
██║  ██╗@
╚═╝  ╚═╝@
        @@
██╗     @
██║     @
██║     @
██║     @
███████╗@
╚══════╝@
        @@
███╗   ███╗@
████╗ ████║@
██╔████╔██║@
██║╚██╔╝██║@
██║ ╚═╝ ██║@
╚═╝     ╚═╝@
           @@
███╗   ██╗@
████╗  ██║@
██╔██╗ ██║@
██║╚██╗██║@
██║ ╚████║@
╚═╝  ╚═══╝@
          @@
 ██████╗ @
██╔═══██╗@
██║   ██║@
██║   ██║@
╚██████╔╝@
 ╚═════╝ @
         @@
██████╗ @
██╔══██╗@
██████╔╝@
██╔═══╝ @
██║     @
╚═╝     @
        @@
 ██████╗ @
██╔═══██╗@
██║   ██║@
██║▄▄ ██║@
╚██████╔╝@
 ╚══▀▀═╝ @
         @@
██████╗ @
██╔══██╗@
██████╔╝@
██╔══██╗@
██║  ██║@
╚═╝  ╚═╝@
        @@
███████╗@
██╔════╝@
███████╗@
╚════██║@
███████║@
╚══════╝@
        @@
████████╗@
╚══██╔══╝@
   ██║   @
   ██║   @
   ██║   @
   ╚═╝   @
         @@
██╗   ██╗@
██║   ██║@
██║   ██║@
██║   ██║@
╚██████╔╝@
 ╚═════╝ @
         @@
██╗   ██╗@
██║   ██║@
██║   ██║@
╚██╗ ██╔╝@
 ╚████╔╝ @
  ╚═══╝  @
         @@
██╗    ██╗@
██║    ██║@
██║ █╗ ██║@
██║███╗██║@
╚███╔███╔╝@
 ╚══╝╚══╝ @
          @@
██╗  ██╗@
╚██╗██╔╝@
 ╚███╔╝ @
 ██╔██╗ @
██╔╝ ██╗@
╚═╝  ╚═╝@
        @@
██╗   ██╗@
╚██╗ ██╔╝@
 ╚████╔╝ @
  ╚██╔╝  @
   ██║   @
   ╚═╝   @
         @@
███████╗@
╚══███╔╝@
  ███╔╝ @
 ███╔╝  @
███████╗@
╚══════╝@
        @@
@
@
@
@
@
@
@@
@
@
@
@
@
@
@@
@
@
@
@
@
@
@@
@
@
@
@
@
@
@@
@
@
@
@
@
@
@@
@
@
@
@
@
@
@@
@
@
@
@
@
@
@@
@
@
@
@
@
@
@@
@
@
@
@
@
@
@@
@
@
@
@
@
@
@@
@
@
@
@
@
@
@@"""


# === Font Management ===

_default_font = None
_builtin_fonts = {}


def _get_builtin_font():
    """Load the built-in ansi-shadow font from embedded data."""
    global _default_font
    if _default_font is None:
        try:
            _default_font = BigFont(ANSI_SHADOW_FONT_DATA.split('\n'), name='ansi-shadow')
        except Exception:
            logging.warning("Failed to load built-in ansi-shadow font", exc_info=1)
    return _default_font


def render(text, font=None):
    """Render text in big font and return as a string.

    Uses given BigFont if specified, otherwise the built-in ansi-shadow font."""
    if font is None:
        font = _get_builtin_font()
    
    if font is None:
        return None
    
    return font.render(text)


# === Public API ===

def terminascii(text: str) -> Optional[str]:
    """
    Generate ASCII art banner from text using the ansi-shadow font.

    This function creates ASCII art text that can be used for titles,
    banners, or any decorative text display in terminal applications.

    Args:
        text: The text to convert to ASCII art

    Returns:
        ASCII art string with trailing whitespace removed, or None if generation fails

    Example:
        ```python
        from terminaide import terminascii

        banner = terminascii("HELLO")
        if banner:
            print(banner)
        ```
    """
    if not text:
        return None

    try:
        # Generate ASCII art using built-in font
        big_letter_obj = render(text)
        if big_letter_obj is None:
            logger.error("font render returned None")
            return None
        ascii_text = str(big_letter_obj)

        logger.debug(f"Generated ASCII art for text: {text}")

        # Remove ALL trailing whitespace and newlines
        ascii_text = ascii_text.rstrip()

        return ascii_text

    except Exception as e:
        logger.warning(f"Failed to generate ASCII art: {e}")
        return None


# data_models.py

"""Defines Pydantic-based settings for terminaide, including path handling for root/non-root mounting and multiple script routing."""

import os
import sys
import logging
from pathlib import Path
from typing import Dict, Any, Optional, Union, List, Callable
from pydantic import BaseModel, Field, field_validator, model_validator

from .exceptions import ConfigurationError

logger = logging.getLogger("terminaide")


class TTYDOptions(BaseModel):
    """TTYd-specific options like auth, interface, and client capacity."""

    writable: bool = True
    port: int = Field(default=7681, gt=1024, lt=65535)
    interface: str = "0.0.0.0"  # Changed from "127.0.0.1" to bind to all interfaces
    check_origin: bool = True
    max_clients: int = Field(default=1, gt=0)
    credential_required: bool = False
    username: Optional[str] = None
    password: Optional[str] = None
    force_https: bool = False

    @model_validator(mode="after")
    def validate_credentials(self) -> "TTYDOptions":
        """Require username/password if credentials are enabled."""
        if self.credential_required and not (self.username and self.password):
            raise ConfigurationError(
                "Both username and password must be provided when credential_required=True"
            )
        return self


class ThemeConfig(BaseModel):
    """Defines basic color and font options for the terminal."""

    background: str = "black"
    foreground: str = "white"
    cursor: str = "white"
    cursor_accent: Optional[str] = None
    selection: Optional[str] = None
    font_family: Optional[str] = None
    font_size: Optional[int] = Field(default=None, gt=0)


class RouteConfigBase(BaseModel):
    """Base class for all route configurations."""

    route_path: str
    preview_image: Optional[Path] = None
    title: Optional[str] = None

    @field_validator("route_path")
    @classmethod
    def validate_route_path(cls, v: str) -> str:
        """Normalize route path to start with '/' and remove trailing '/'."""
        if not v.startswith("/"):
            v = f"/{v}"
        if v != "/" and v.endswith("/"):
            v = v.rstrip("/")
        return v

    @field_validator("preview_image")
    @classmethod
    def validate_preview_image_path(
        cls, v: Optional[Union[str, Path]]
    ) -> Optional[Path]:
        """
        Ensure the preview image file exists if provided, trying:
        1. The path as provided (relative to CWD or absolute)
        2. The path relative to the main script being executed
        3. Look in common static dirs relative to main script (static, assets, images)
        4. Look relative to project root directories
        """
        if v is None:
            return None

        original_path = Path(v)

        # Strategy 1: Use the path as-is (absolute or relative to CWD)
        if original_path.is_absolute() or original_path.exists():
            return original_path.absolute()

        # Strategy 2: Try relative to the main script being run
        try:
            main_script = Path(sys.argv[0]).absolute()
            main_script_dir = main_script.parent
            image_relative_path = main_script_dir / original_path
            if image_relative_path.exists():
                logger.debug(
                    f"Found preview image at {image_relative_path} (relative to main script)"
                )
                return image_relative_path.absolute()

            # Strategy 3: Try common static directories relative to the main script
            for common_dir in ["static", "assets", "images", "img"]:
                common_path = main_script_dir / common_dir / original_path.name
                if common_path.exists():
                    logger.debug(
                        f"Found preview image at {common_path} (in common directory)"
                    )
                    return common_path.absolute()

            # Strategy 4: Try going up directories to find project root with static dirs
            current_dir = main_script_dir
            for _ in range(3):  # Limit depth to prevent excessive searching
                parent_dir = current_dir.parent
                if parent_dir == current_dir:  # Reached filesystem root
                    break

                current_dir = parent_dir
                for common_dir in ["static", "assets", "images", "img"]:
                    common_path = current_dir / common_dir / original_path.name
                    if common_path.exists():
                        logger.debug(
                            f"Found preview image at {common_path} (in project structure)"
                        )
                        return common_path.absolute()
        except Exception as e:
            logger.debug(
                f"Error resolving preview image path relative to main script: {e}"
            )

        # If we got here, log a warning but don't fail - we'll fall back to the default
        logger.warning(
            f"Preview image does not exist: {v}. Will use default preview image."
        )
        return None

    def is_terminal_route(self) -> bool:
        """Check if this route requires a terminal (ttyd process)."""
        return False


class ScriptConfig(RouteConfigBase):
    """Configuration for a single terminal route, including the script path, port assignment, and optional custom title."""

    script: Optional[Path] = None
    args: List[str] = Field(default_factory=list)
    port: Optional[int] = None
    function_object: Optional[Callable] = None
    _function_wrapper_path: Optional[Path] = None
    dynamic: bool = False  # Enable dynamic argument passing via query parameters
    args_param: str = "args"  # Query parameter name for dynamic arguments
    _dynamic_wrapper_path: Optional[Path] = None

    @field_validator("script")
    @classmethod
    def validate_script_path(cls, v: Optional[Union[str, Path]]) -> Optional[Path]:
        """
        Ensure the script file exists, trying:
        1. The path as provided (relative to CWD or absolute)
        2. The path relative to the main script being executed

        Returns None if no script path is provided (function-based route).
        """
        if v is None:
            return None

        original_path = Path(v)

        # Strategy 1: Use the path as-is (absolute or relative to CWD)
        if original_path.is_absolute() or original_path.exists():
            return original_path.absolute()

        # Strategy 2: Try relative to the main script being run
        try:
            main_script = Path(sys.argv[0]).absolute()
            main_script_dir = main_script.parent
            script_relative_path = main_script_dir / original_path
            if script_relative_path.exists():
                logger.debug(
                    f"Found script at {script_relative_path} (relative to main script)"
                )
                return script_relative_path.absolute()
        except Exception as e:
            logger.debug(f"Error resolving path relative to main script: {e}")

        # If we got here, the path doesn't exist
        error_msg = f"Script file does not exist: {v}\n"

        # Add context about where we looked
        cwd_path = Path.cwd() / v
        error_msg += f"Current working directory: {os.getcwd()}\n"
        error_msg += f"Tried:\n"
        error_msg += f"  - As provided: {v}\n"
        error_msg += f"  - Relative to CWD: {cwd_path}\n"

        # Add info about script-relative path if available
        if sys.argv and len(sys.argv) > 0:
            script_path = Path(sys.argv[0]).absolute().parent / v
            error_msg += f"  - Relative to main script: {script_path}\n"

        raise ConfigurationError(error_msg)

    @field_validator("args")
    @classmethod
    def validate_args(cls, v: List[str]) -> List[str]:
        """Convert all args to strings."""
        return [str(arg) for arg in v]

    @model_validator(mode="after")
    def validate_script_or_function(self) -> "ScriptConfig":
        """Ensure either script or function_object is provided."""
        if self.script is None and self.function_object is None:
            raise ConfigurationError(
                "Either script or function_object must be provided for route "
                + f"'{self.route_path}'"
            )
        return self

    @property
    def is_function_based(self) -> bool:
        """Returns True if this route uses a function instead of a script."""
        return self.function_object is not None

    def set_function_wrapper_path(self, path: Path) -> None:
        """Set the path to the generated wrapper script for a function."""
        self._function_wrapper_path = path
    
    def set_dynamic_wrapper_path(self, path: Path) -> None:
        """Set the path to the generated dynamic wrapper script."""
        self._dynamic_wrapper_path = path

    @property
    def effective_script_path(self) -> Path:
        """Returns the script path to use (priority: dynamic_wrapper > function_wrapper > script)."""
        if self.dynamic and self._dynamic_wrapper_path is not None:
            return self._dynamic_wrapper_path
        if self.is_function_based and self._function_wrapper_path is not None:
            return self._function_wrapper_path
        return self.script

    def is_terminal_route(self) -> bool:
        """ScriptConfig routes always require a terminal."""
        return True


class IndexPageConfig(RouteConfigBase):
    """Configuration for an index page route."""

    index_page: Any  # Will be IndexPage instance, using Any to avoid circular import

    def is_terminal_route(self) -> bool:
        """Index pages don't require terminals."""
        return False

    def get_preview_image(self) -> Optional[Path]:
        """Get preview image from either the config or the IndexPage."""
        if self.preview_image:
            return self.preview_image
        elif hasattr(self.index_page, "preview_image"):
            return self.index_page.preview_image
        return None


class TTYDConfig(BaseModel):
    """Main configuration for terminaide, handling root vs. non-root mounting, multiple scripts, and other settings like theme and debug mode."""

    script: Optional[Path] = None
    mount_path: str = "/"
    port: int = Field(default=7681, gt=1024, lt=65535)
    theme: ThemeConfig = Field(default_factory=ThemeConfig)
    ttyd_options: TTYDOptions = Field(default_factory=TTYDOptions)
    template_override: Optional[Path] = None
    preview_image: Optional[Path] = None  # Added preview_image field
    log_level: Optional[str] = "info"  # "debug", "info", "warning", "error", None
    title: str = "Terminal"
    route_configs: List[Union[ScriptConfig, IndexPageConfig]] = Field(
        default_factory=list
    )
    _mode: str = "script"  # Default mode: "function", "script", "apps", or "meta"
    forward_env: Union[bool, List[str], Dict[str, Optional[str]]] = True
    venv_detection: bool = True  # Enable automatic virtual environment detection

    # Legacy field names for backward compatibility
    @property
    def script_configs(self) -> List[ScriptConfig]:
        """Backward compatibility: return only ScriptConfig instances."""
        return [cfg for cfg in self.route_configs if isinstance(cfg, ScriptConfig)]

    @field_validator("script", "template_override")
    @classmethod
    def validate_paths(cls, v: Optional[Union[str, Path]]) -> Optional[Path]:
        """Ensure given path exists, if provided."""
        if v is None:
            return None
        path = Path(v)
        if not path.exists():
            raise ConfigurationError(f"Path does not exist: {path}")
        return path.absolute()

    @field_validator("preview_image")
    @classmethod
    def validate_preview_image(cls, v: Optional[Union[str, Path]]) -> Optional[Path]:
        """Ensure preview image exists, if provided."""
        if v is None:
            return None
        path = Path(v)
        if not path.exists():
            logger.warning(
                f"Preview image path does not exist: {path}. Will use default preview image."
            )
            return None
        return path.absolute()

    @field_validator("mount_path")
    @classmethod
    def validate_mount_path(cls, v: str) -> str:
        """Normalize and disallow '/terminal' as a mount path."""
        if v in ("", "/"):
            return "/"
        if not v.startswith("/"):
            v = f"/{v}"
        v = v.rstrip("/")
        if v == "/terminal":
            raise ConfigurationError(
                '"/terminal" is reserved. Please use another mount path.'
            )
        return v

    @model_validator(mode="after")
    def validate_route_configs(self) -> "TTYDConfig":
        """Check for unique route paths and handle a default script if no routes given."""
        seen_routes = set()
        for config in self.route_configs:
            if config.route_path in seen_routes:
                raise ConfigurationError(f"Duplicate route path: {config.route_path}")
            seen_routes.add(config.route_path)

        # Add a default script config if needed (backward compatibility)
        if not self.route_configs and self.script:
            self.route_configs.append(
                ScriptConfig(
                    route_path="/",
                    script=self.script,
                    args=[],
                    port=self.port,
                    title=self.title,
                    preview_image=self.preview_image,
                )
            )
        return self

    @property
    def is_root_mounted(self) -> bool:
        """True if mounted at root ('/')."""
        return self.mount_path == "/"

    @property
    def is_multi_script(self) -> bool:
        """True if multiple scripts are configured."""
        return len(self.script_configs) > 1

    @property
    def has_index_pages(self) -> bool:
        """True if any index pages are configured."""
        return any(isinstance(cfg, IndexPageConfig) for cfg in self.route_configs)

    @property
    def is_meta_mode(self) -> bool:
        """True if this is a meta-server configuration."""
        return self._mode == "meta"

    @property
    def terminal_path(self) -> str:
        """Return the terminal's path, accounting for root or non-root mounting."""
        if self.is_root_mounted:
            return "/terminal"
        return f"{self.mount_path}/terminal"

    @property
    def static_path(self) -> str:
        """Return the path for static files."""
        if self.is_root_mounted:
            return "/static"
        return f"{self.mount_path}/static"

    def get_route_config_for_path(
        self, path: str
    ) -> Optional[Union[ScriptConfig, IndexPageConfig]]:
        """
        Find which route config matches an incoming request path.
        """
        # For single route configs, return it if the path matches
        if len(self.route_configs) == 1:
            return self.route_configs[0]

        # Sort by path length to match most specific first
        sorted_configs = sorted(
            self.route_configs, key=lambda c: len(c.route_path), reverse=True
        )

        for config in sorted_configs:
            if path == config.route_path or path.startswith(config.route_path + "/"):
                return config

        return None

    def get_script_config_for_path(self, path: str) -> Optional[ScriptConfig]:
        """
        Find which script config matches an incoming request path,
        returning the default if none match.
        """
        config = self.get_route_config_for_path(path)
        if config and isinstance(config, ScriptConfig):
            return config

        # Fallback to first script config if available
        for cfg in self.route_configs:
            if isinstance(cfg, ScriptConfig):
                return cfg

        return None

    def get_terminal_path_for_route(self, route_path: str) -> str:
        """Return the terminal path for a specific route, or global path if root."""
        if route_path == "/":
            return self.terminal_path
        return f"{route_path}/terminal"

    def get_health_check_info(self) -> Dict[str, Any]:
        """Return structured data about the config for health checks."""
        route_info = []

        for config in self.route_configs:
            if isinstance(config, ScriptConfig):
                route_info.append(
                    {
                        "type": "terminal",
                        "route_path": config.route_path,
                        "script": str(config.effective_script_path),
                        "is_function": config.is_function_based,
                        "args": config.args,
                        "port": config.port,
                        "title": config.title or self.title,
                        "preview_image": (
                            str(config.preview_image) if config.preview_image else None
                        ),
                    }
                )
            elif isinstance(config, IndexPageConfig):
                route_info.append(
                    {
                        "type": "index",
                        "route_path": config.route_path,
                        "title": config.title
                        or getattr(config.index_page, "page_title", "Index"),
                        "preview_image": (
                            str(config.get_preview_image())
                            if config.get_preview_image()
                            else None
                        ),
                        "menu_items": len(config.index_page.get_all_menu_items()),
                    }
                )

        # Include meta-server specific info if in meta mode
        meta_info = {}
        if self.is_meta_mode:
            meta_info = {
                "is_meta_server": True,
                "app_dir": str(getattr(self, "_app_dir", "auto-detected")),
            }

        return {
            "mount_path": self.mount_path,
            "terminal_path": self.terminal_path,
            "static_path": self.static_path,
            "is_root_mounted": self.is_root_mounted,
            "is_multi_script": self.is_multi_script,
            "has_index_pages": self.has_index_pages,
            "entry_mode": self._mode,  # Add entry mode to health check info
            "port": self.port,
            "log_level": self.log_level,
            "max_clients": self.ttyd_options.max_clients,
            "auth_required": self.ttyd_options.credential_required,
            "preview_image": str(self.preview_image) if self.preview_image else None,
            "route_configs": route_info,
            **meta_info,  # Include meta-server info if applicable
        }


def create_route_configs(
    terminal_routes: Dict[str, Union[str, Path, List, Dict[str, Any], Callable]],
) -> List[Union[ScriptConfig, IndexPageConfig]]:
    """
    Convert the terminal_routes dictionary into a list of route configurations.

    Now supports:
    - Script paths (str/Path)
    - Functions (Callable)
    - Lists with script path and args
    - Dictionaries with advanced config for scripts or functions
    - IndexPage instances
    """
    # Import here to avoid circular import
    from .index import AutoIndex

    route_configs = []

    for route_path, route_spec in terminal_routes.items():
        # Handle AutoIndex instances
        if isinstance(route_spec, AutoIndex):
            route_configs.append(
                IndexPageConfig(
                    route_path=route_path,
                    index_page=route_spec,
                    preview_image=route_spec.preview_image,
                    title=route_spec.title,
                )
            )
            continue

        # Handle direct callable function
        if callable(route_spec):
            func = route_spec
            func_name = getattr(func, "__name__", "function")

            route_configs.append(
                ScriptConfig(
                    route_path=route_path,
                    function_object=func,
                    script=None,
                    args=[],
                    title=f"{func_name}()",
                )
            )
            continue

        # Handle dictionary configuration that might contain a function
        if isinstance(route_spec, dict) and "function" in route_spec:
            func = route_spec["function"]
            if not callable(func):
                raise ConfigurationError(
                    f"'function' value for route {route_path} is not callable"
                )

            func_name = getattr(func, "__name__", "function")

            cfg_data = {
                "route_path": route_path,
                "function_object": func,
                "client_script": None,
                "args": [],
            }

            # Use provided title or auto-generate
            if "title" in route_spec:
                cfg_data["title"] = route_spec["title"]
            else:
                cfg_data["title"] = f"{func_name}()"

            if "port" in route_spec:
                cfg_data["port"] = route_spec["port"]

            if "preview_image" in route_spec:
                cfg_data["preview_image"] = route_spec["preview_image"]
            
            if "dynamic" in route_spec:
                cfg_data["dynamic"] = route_spec["dynamic"]
            
            if "args_param" in route_spec:
                cfg_data["args_param"] = route_spec["args_param"]

            route_configs.append(ScriptConfig(**cfg_data))
            continue

        # Handle script path or function in dictionary
        # Support both "script" (new) and "client_script" (backward compatibility)
        if isinstance(route_spec, dict) and ("script" in route_spec or "client_script" in route_spec):
            # Use "script" if present, otherwise fall back to "client_script"
            script_value = route_spec.get("script") or route_spec.get("client_script")
            
            # Check if script is actually a function
            if callable(script_value):
                func = script_value
                func_name = getattr(func, "__name__", "function")
                
                cfg_data = {
                    "route_path": route_path,
                    "function_object": func,
                    "script": None,
                    "args": [],
                }
                
                # Use provided title or auto-generate
                if "title" in route_spec:
                    cfg_data["title"] = route_spec["title"]
                else:
                    cfg_data["title"] = f"{func_name}()"
                
                if "port" in route_spec:
                    cfg_data["port"] = route_spec["port"]
                
                if "preview_image" in route_spec:
                    cfg_data["preview_image"] = route_spec["preview_image"]
                
                if "dynamic" in route_spec:
                    cfg_data["dynamic"] = route_spec["dynamic"]
                
                if "args_param" in route_spec:
                    cfg_data["args_param"] = route_spec["args_param"]
                
                route_configs.append(ScriptConfig(**cfg_data))
                continue
            
            # Handle script path (existing logic)
            if isinstance(script_value, list) and len(script_value) > 0:
                script_path = script_value[0]
                args = script_value[1:]
            else:
                script_path = script_value
                args = []

            if "args" in route_spec:
                args = route_spec["args"]

            cfg_data = {
                "route_path": route_path,
                "script": script_path,
                "args": args,
            }

            # Use provided title or auto-generate if not present
            if "title" in route_spec:
                cfg_data["title"] = route_spec["title"]
            else:
                # Auto-generate title based on script name
                script_name = Path(script_path).name
                cfg_data["title"] = f"{script_name}"

            if "port" in route_spec:
                cfg_data["port"] = route_spec["port"]

            # Handle preview_image if provided in the script_spec
            if "preview_image" in route_spec:
                cfg_data["preview_image"] = route_spec["preview_image"]
            
            if "dynamic" in route_spec:
                cfg_data["dynamic"] = route_spec["dynamic"]
            
            if "args_param" in route_spec:
                cfg_data["args_param"] = route_spec["args_param"]

            route_configs.append(ScriptConfig(**cfg_data))
            continue

        # Handle script path with args as list
        if isinstance(route_spec, list) and len(route_spec) > 0:
            script_path = route_spec[0]
            args = route_spec[1:]

            # Auto-generate title based on script name
            script_name = Path(script_path).name

            route_configs.append(
                ScriptConfig(
                    route_path=route_path,
                    script=script_path,
                    args=args,
                    title=f"{script_name}",
                )
            )
            continue

        # Handle simple script path (str/Path)
        script_path = route_spec

        # Auto-generate title based on script name
        script_name = Path(script_path).name

        route_configs.append(
            ScriptConfig(
                route_path=route_path,
                script=script_path,
                args=[],
                title=f"{script_name}",
            )
        )

    if not route_configs:
        raise ConfigurationError("No valid route configuration provided")

    return route_configs


# Legacy function name for backward compatibility
def create_script_configs(terminal_routes: Dict[str, Any]) -> List[ScriptConfig]:
    """
    Backward compatibility wrapper for create_route_configs.

    Note: This will filter out IndexPage configs and only return ScriptConfig instances.
    """
    all_configs = create_route_configs(terminal_routes)
    return [cfg for cfg in all_configs if isinstance(cfg, ScriptConfig)]



# config.py

"""Core configuration module for Terminaide.

This module contains shared configuration classes and utilities used by different parts of the Terminaide library. It serves as a central point of configuration to avoid circular dependencies.
"""

import sys
import shutil
import logging
import hashlib
from pathlib import Path
from dataclasses import dataclass, field
from contextlib import asynccontextmanager
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from fastapi import FastAPI, Request, WebSocket
from typing import Optional, Dict, Union, Tuple, List, Callable, Any

from .proxy import ProxyManager
from .terminal import TTYDManager
from .exceptions import TemplateError
from .models import (
    TTYDConfig,
    ThemeConfig,
    TTYDOptions,
    ScriptConfig,
    IndexPageConfig,
    create_route_configs,
)

logger = logging.getLogger("terminaide")


def smart_resolve_path(path: Union[str, Path, Callable]) -> Union[Path, Callable]:
    """Resolves a path using a predictable strategy:
    1. If it's a callable function, return it as-is
    2. First try the path as-is (absolute or relative to CWD)
    3. Then try relative to the main script being run (sys.argv[0])

    This approach is both flexible and predictable.
    """
    # If path is a callable function, return it directly
    if callable(path):
        return path

    original_path = Path(path)

    # Strategy 1: Use the path as-is (absolute or relative to CWD)
    if original_path.is_absolute() or original_path.exists():
        return original_path.absolute()

    # Strategy 2: Try relative to the main script being run
    try:
        main_script = Path(sys.argv[0]).absolute()
        main_script_dir = main_script.parent
        script_relative_path = main_script_dir / original_path
        if script_relative_path.exists():
            logger.debug(
                f"Found script at {script_relative_path} (relative to main script)"
            )
            return script_relative_path.absolute()
    except Exception as e:
        logger.debug(f"Error resolving path relative to main script: {e}")

    # Return the original if nothing was found
    return original_path


def copy_preview_image_to_static(preview_image: Path) -> str:
    """
    Copy a preview image to the cache assets directory and then to static directory.
    Returns the filename of the copied image in the static directory.
    """
    logger.debug(f"copy_preview_image_to_static called with: {preview_image}")

    if not preview_image or not preview_image.exists():
        logger.debug(f"Preview image doesn't exist, using default: {preview_image}")
        return "preview.png"  # Use default

    # Get both directories
    package_dir = Path(__file__).parent.parent
    assets_dir = package_dir / "cache" / "assets"
    static_dir = package_dir / "static"
    
    # Ensure both directories exist
    assets_dir.mkdir(exist_ok=True, parents=True)
    static_dir.mkdir(exist_ok=True)

    # Generate a unique filename based on the image content hash
    try:
        # Generate a hash of the file content to create a unique name
        file_hash = hashlib.md5(preview_image.read_bytes()).hexdigest()[:12]
        extension = preview_image.suffix.lower()

        # Only allow common image extensions
        if extension not in (".png", ".jpg", ".jpeg", ".gif", ".svg"):
            logger.warning(
                f"Unsupported image extension: {extension}. Using default preview."
            )
            return "preview.png"

        new_filename = f"preview_{file_hash}{extension}"
        assets_path = assets_dir / new_filename
        static_path = static_dir / new_filename

        # Copy to assets directory first (source of truth)
        shutil.copy2(preview_image, assets_path)
        logger.debug(f"Copied preview image from {preview_image} to {assets_path}")
        
        # Then copy to static directory for serving
        shutil.copy2(assets_path, static_path)
        logger.debug(f"Copied preview image from {assets_path} to {static_path}")

        return new_filename
    except Exception as e:
        logger.warning(f"Failed to copy preview image: {e}. Using default preview.")
        return "preview.png"


@dataclass
class TerminaideConfig:
    """Unified configuration for all Terminaide serving modes."""

    # Common configuration options
    port: int = 8000
    title: str = "Terminal"
    theme: Dict[str, Any] = field(
        default_factory=lambda: {"background": "black", "foreground": "white"}
    )
    log_level: Optional[str] = "info"  # "debug", "info", "warning", "error", None
    banner: Union[bool, str] = True
    forward_env: Union[bool, List[str], Dict[str, Optional[str]]] = True

    # Advanced configuration
    ttyd_options: Dict[str, Any] = field(default_factory=dict)
    template_override: Optional[Path] = None
    trust_proxy_headers: bool = True
    mount_path: str = "/"

    # Preview image configuration
    preview_image: Optional[Path] = None

    # Script/function arguments
    args: Optional[List[str]] = None
    dynamic: bool = False
    args_param: str = "args"

    # Proxy settings
    ttyd_port: int = 7681  # Base port for ttyd processes
    
    # Cache configuration
    ephemeral_cache_dir: Optional[Path] = None  # Override for ephemeral script storage
    monitor_log_path: Optional[Path] = None  # Override for monitor log file location

    # Internal fields (not exposed directly)
    _target: Optional[Union[Callable, Path, Dict[str, Any]]] = None
    _app: Optional[FastAPI] = None
    _mode: str = "function"  # "function", "script", "apps", or "meta"


def build_config(
    config: Optional[TerminaideConfig], overrides: Dict[str, Any]
) -> TerminaideConfig:
    """Build a config object from the provided config and overrides."""
    if config is None:
        config = TerminaideConfig()

    # Handle backward compatibility for deprecated parameters
    if "debug" in overrides:
        logger.warning("'debug' parameter is deprecated. Use log_level='debug' instead.")
        if overrides["debug"]:
            overrides["log_level"] = "debug"
        del overrides["debug"]
    
    if "configure_logging" in overrides:
        logger.warning("'configure_logging' parameter is deprecated. Use log_level=None instead.")
        if not overrides["configure_logging"]:
            overrides["log_level"] = None
        del overrides["configure_logging"]

    # Apply overrides
    for key, value in overrides.items():
        if hasattr(config, key):
            setattr(config, key, value)

    # Validate log_level
    if config.log_level is not None:
        valid_levels = ["debug", "info", "warning", "error", "critical"]
        if config.log_level.lower() not in valid_levels:
            raise ValueError(f"Invalid log_level: {config.log_level}. Must be one of {valid_levels} or None")
        config.log_level = config.log_level.lower()

    return config


def setup_templates(config: TerminaideConfig) -> Tuple[Jinja2Templates, str]:
    """Set up the Jinja2 templates for the HTML interface."""
    if config.template_override:
        template_dir = config.template_override.parent
        template_file = config.template_override.name
    else:
        template_dir = Path(__file__).parent.parent / "templates"
        template_file = "terminal.html"

    if not template_dir.exists():
        raise TemplateError(str(template_dir), "Template directory not found")

    templates = Jinja2Templates(directory=str(template_dir))

    if not (template_dir / template_file).exists():
        raise TemplateError(template_file, "Template file not found")

    return templates, template_file


def configure_routes(
    app: FastAPI,
    config: TTYDConfig,
    ttyd_manager: TTYDManager,
    proxy_manager: ProxyManager,
    templates: Jinja2Templates,
    template_file: str,
) -> None:
    """Define routes for TTYD terminals and index pages."""

    # Check if index template exists
    template_dir = Path(templates.env.loader.searchpath[0])
    index_template_file = "index.html"
    has_index_template = (template_dir / index_template_file).exists()

    if config.has_index_pages and not has_index_template:
        logger.warning(
            f"Index pages configured but {index_template_file} not found in templates directory"
        )

    @app.get(f"{config.mount_path}/health")
    async def health_check():
        return {
            "ttyd": ttyd_manager.check_health(),
            "proxy": proxy_manager.get_routes_info(),
        }

    # Process all route configs
    for route_config in config.route_configs:
        route_path = route_config.route_path

        if isinstance(route_config, IndexPageConfig):
            # Handle index page route
            @app.get(route_path, response_class=HTMLResponse)
            async def index_page_handler(
                request: Request,
                route_config=route_config,
            ):
                try:
                    # Get the index page instance
                    index_page = route_config.index_page

                    # Build template context
                    context = index_page.to_template_context()
                    context["request"] = request

                    # Handle preview image
                    preview_image_path = route_config.get_preview_image()
                    if preview_image_path and preview_image_path.exists():
                        preview_image = copy_preview_image_to_static(preview_image_path)
                    else:
                        preview_image = "preview.png"

                    context["preview_image"] = preview_image

                    logger.debug(f"Rendering index page for route {route_path}")

                    # Check if index template exists
                    if not has_index_template:
                        return HTMLResponse(
                            content=f"<h1>Index template not found</h1><p>Please create {index_template_file}</p>",
                            status_code=500,
                        )

                    return templates.TemplateResponse(index_template_file, context)
                except Exception as e:
                    logger.error(
                        f"Error rendering index page for route {route_path}: {e}"
                    )
                    raise TemplateError(index_template_file, str(e))

        elif isinstance(route_config, ScriptConfig):
            # Handle terminal route
            terminal_path = config.get_terminal_path_for_route(route_path)
            title = route_config.title or config.title

            # Debug logging for preview image config
            logger.debug(f"Script config preview_image: {route_config.preview_image}")
            logger.debug(f"Config preview_image: {config.preview_image}")

            # Get preview image path - prefer script_config's image, fall back to config's, then default
            preview_image = "preview.png"  # Default fallback

            # Try script config preview first
            if route_config.preview_image and route_config.preview_image.exists():
                logger.debug(
                    f"Using script config preview image: {route_config.preview_image}"
                )
                preview_image = copy_preview_image_to_static(route_config.preview_image)
            elif route_config.preview_image:
                logger.warning(
                    f"Script preview image doesn't exist: {route_config.preview_image}"
                )

            # If no script preview or it failed, try global config
            elif config.preview_image:
                logger.debug(
                    f"Using global config preview image: {config.preview_image}"
                )
                if config.preview_image.exists():
                    preview_image = copy_preview_image_to_static(config.preview_image)
                else:
                    logger.warning(
                        f"Global preview image doesn't exist: {config.preview_image}"
                    )
            else:
                logger.debug("No custom preview images configured, using default")

            logger.debug(f"Final preview image for route {route_path}: {preview_image}")

            @app.get(route_path, response_class=HTMLResponse)
            async def terminal_interface(
                request: Request,
                route_path=route_path,
                terminal_path=terminal_path,
                title=title,
                preview_image=preview_image,
                route_config=route_config,
            ):
                try:
                    logger.debug(
                        f"Rendering template with preview_image={preview_image}"
                    )
                    
                    # For dynamic routes, append query parameters to the terminal path
                    iframe_src = terminal_path
                    if route_config.dynamic and request.url.query:
                        iframe_src = f"{terminal_path}?{request.url.query}"
                        logger.debug(f"Dynamic route {route_path}: appending query params to iframe src: {iframe_src}")
                    
                    return templates.TemplateResponse(
                        template_file,
                        {
                            "request": request,
                            "mount_path": iframe_src,
                            "theme": config.theme.model_dump(),
                            "title": title,
                            "preview_image": preview_image,
                        },
                    )
                except Exception as e:
                    logger.error(
                        f"Template rendering error for route {route_path}: {e}"
                    )
                    raise TemplateError(template_file, str(e))

            @app.websocket(f"{terminal_path}/ws")
            async def terminal_ws(websocket: WebSocket, route_path=route_path):
                await proxy_manager.proxy_websocket(websocket, route_path=route_path)

            @app.api_route(
                f"{terminal_path}/{{path:path}}",
                methods=[
                    "GET",
                    "POST",
                    "PUT",
                    "DELETE",
                    "OPTIONS",
                    "HEAD",
                    "PATCH",
                    "TRACE",
                ],
            )
            async def proxy_terminal_request(
                request: Request, path: str, route_path=route_path
            ):
                return await proxy_manager.proxy_http(request)


def configure_app(app: FastAPI, config: TTYDConfig):
    """Configure the FastAPI app with the TTYDManager, ProxyManager, and routes."""
    mode = "apps-server" if config.is_multi_script else "solo-server"
    entry_mode = getattr(config, "_mode", "script")

    # Debug-level configuration logging to avoid duplication
    if entry_mode == "meta":
        logger.debug(f"Configuring meta-server with {config.mount_path} mounting")
    else:
        logger.debug(
            f"Configuring {mode} with {config.mount_path} mounting ({entry_mode} API)"
        )

    ttyd_manager = TTYDManager(config)
    proxy_manager = ProxyManager(config)

    package_dir = Path(__file__).parent.parent
    static_dir = package_dir / "static"
    static_dir.mkdir(exist_ok=True)

    terminaide_static_path = "/terminaide-static"  # one-place constant
    app.mount(
        terminaide_static_path,
        StaticFiles(directory=str(static_dir)),
        name="terminaide_static",  # unique route-name
    )

    # config.static_path = terminaide_static_path

    templates, template_file = setup_templates(config)
    app.state.terminaide_templates = templates
    app.state.terminaide_template_file = template_file
    app.state.terminaide_config = config

    configure_routes(app, config, ttyd_manager, proxy_manager, templates, template_file)

    return ttyd_manager, proxy_manager


@asynccontextmanager
async def terminaide_lifespan(app: FastAPI, config: TTYDConfig):
    """Lifespan context manager for the TTYDManager and ProxyManager."""
    ttyd_manager, proxy_manager = configure_app(app, config)

    mode = "apps-server" if config.is_multi_script else "solo-server"
    entry_mode = getattr(config, "_mode", "script")

    # Consolidated startup log message
    mode_desc = "meta-server" if entry_mode == "meta" else f"{mode} ({entry_mode} API)"
    logger.info(
        f"Starting ttyd service in {mode_desc} mode "
        f"(mounting: {'root' if config.is_root_mounted else 'non-root'})"
    )

    ttyd_manager.start()
    try:
        yield
    finally:
        logger.info("Cleaning up ttyd service...")
        
        # Clean up ephemeral files on graceful shutdown
        try:
            from .wrappers import cleanup_own_ephemeral_files
            cleanup_own_ephemeral_files()
        except ImportError:
            pass  # Graceful fallback if import fails
        
        ttyd_manager.stop()
        await proxy_manager.cleanup()


def convert_terminaide_config_to_ttyd_config(
    config: TerminaideConfig, script_path: Path = None
) -> TTYDConfig:
    """Convert a TerminaideConfig to a TTYDConfig."""
    if (
        script_path is None
        and config._target is not None
        and isinstance(config._target, Path)
    ):
        script_path = config._target

    terminal_routes = {}
    if config._mode == "apps" and isinstance(config._target, dict):
        terminal_routes = config._target
    elif script_path is not None:
        # For script mode, include args and dynamic if provided
        if config.args is not None or config.dynamic:
            terminal_routes = {"/": {
                "script": script_path,
                "args": config.args or [],
                "dynamic": config.dynamic,
                "args_param": config.args_param,
            }}
        else:
            terminal_routes = {"/": script_path}
    elif callable(config._target):
        # Handle function target for serve_function mode
        if config.args is not None or config.dynamic:
            terminal_routes = {"/": {
                "function": config._target,
                "args": config.args or [],
                "dynamic": config.dynamic,
                "args_param": config.args_param,
            }}
        else:
            terminal_routes = {"/": config._target}

    # Use create_route_configs instead of create_script_configs
    route_configs = create_route_configs(terminal_routes)

    # If we have route configs and a custom title is set, apply it to the first script config
    if route_configs and config.title != "Terminal":
        for cfg in route_configs:
            if isinstance(cfg, ScriptConfig):
                cfg.title = config.title
                break

    # Debug log for preview_image
    if hasattr(config, "preview_image") and config.preview_image:
        logger.debug(
            f"Converting preview_image from TerminaideConfig: {config.preview_image}"
        )

    # Convert theme dict to ThemeConfig
    theme_config = ThemeConfig(**(config.theme or {}))

    # Convert ttyd_options dict to TTYDOptions
    ttyd_options_config = TTYDOptions(**(config.ttyd_options or {}))

    # Find the first script config for backward compatibility
    first_script_config = None
    for cfg in route_configs:
        if isinstance(cfg, ScriptConfig):
            first_script_config = cfg
            break

    ttyd_config = TTYDConfig(
        script=(
            first_script_config.script
            if first_script_config and not first_script_config.is_function_based
            else None
        ),
        mount_path=config.mount_path,
        port=config.ttyd_port,
        theme=theme_config,
        ttyd_options=ttyd_options_config,
        template_override=config.template_override,
        preview_image=config.preview_image,  # Pass the preview_image to TTYDConfig
        title=config.title,  # Keep the original title
        log_level=config.log_level,
        route_configs=route_configs,  # Use route_configs instead of script_configs
        forward_env=config.forward_env,
    )

    # Propagate the entry mode to TTYDConfig - include meta mode
    ttyd_config._mode = config._mode

    # Debug log for meta mode
    if config._mode == "meta":
        logger.debug(f"Converting meta-server config to TTYDConfig")
        # Copy any special meta-specific attributes
        if hasattr(config, "_app_dir"):
            ttyd_config._app_dir = config._app_dir

    return ttyd_config



# terminaide/__init__.py

"""terminaide: Serve Python CLI applications in the browser using ttyd.

This package provides tools to easily serve Python CLI applications through
a browser-based terminal using ttyd. It handles binary installation and
management automatically across supported platforms.

Public API:
- serve_function(): Serve a Python function in a browser terminal
- serve_script(): Serve a Python script file in a terminal  
- serve_apps(): Integrate multiple terminals into a FastAPI application
- AutoIndex: Create navigable index pages (HTML or Curses)
- terminascii(): Generate ASCII banners
- Monitor: Process output monitoring with rich terminal interface

Supported Platforms:
- Linux x86_64 (Docker containers)
- macOS ARM64 (Apple Silicon)
"""

import logging
from pathlib import Path
from fastapi import FastAPI
from typing import Optional, Dict, Any, Union, List, Callable

from .core.config import TerminaideConfig, build_config
from .core.factory import ServeWithConfig
from .core.index import AutoIndex
from .core.terminascii import terminascii
from .core.monitor import ServerMonitor

# Get package-level logger (configuration happens when serve_* functions are called)
logger = logging.getLogger("terminaide")

################################################################################
# Shared Helper Functions
################################################################################


def _prepare_config(
    config: Optional[TerminaideConfig], banner: Union[bool, str], **kwargs
) -> TerminaideConfig:
    """Prepare configuration with common parameters."""
    kwargs["banner"] = banner
    return build_config(config, kwargs)


def _auto_generate_title(
    cfg: TerminaideConfig, mode: str, target: Any, kwargs: Dict
) -> None:
    """Auto-generate title if not specified by user."""
    if "title" in kwargs or (cfg.title != "Terminal"):
        return

    if mode == "function":
        cfg.title = f"{target.__name__}()"
    elif mode == "script":
        if hasattr(cfg, "_original_function_name"):
            cfg.title = f"{cfg._original_function_name}()"
        else:
            cfg.title = Path(target).name


################################################################################
# APIs
################################################################################


def serve_script(
    script_path: Union[str, Path],
    port: int = 8000,
    title: Optional[str] = None,
    theme: Optional[Dict[str, str]] = None,
    log_level: Optional[str] = "info",
    args: Optional[List[str]] = None,
    dynamic: bool = False,
    args_param: str = "args",
) -> None:
    """Serve a Python script in a browser terminal.

    Creates a web-accessible terminal that runs the provided Python script.

    Args:
        script_path: Path to the script file to serve
        port: Web server port (default: 8000)
        title: Terminal window title (default: auto-generated from script name)
        theme: Terminal theme colors (default: {"background": "black", "foreground": "white"})
        log_level: Logging level ("debug", "info", "warning", "error", None) (default: "info")
        args: Command-line arguments to pass to the script (default: None)
        dynamic: Enable dynamic arguments via URL query parameters (default: False)
        args_param: Name of the query parameter for dynamic arguments (default: "args")

    Examples:
        Basic usage:
            serve_script("my_script.py")

        With custom configuration:
            serve_script("my_script.py", port=8080, title="My Script")
            serve_script("my_script.py", theme={"background": "navy"})
            serve_script("my_script.py", log_level="debug")
        
        With arguments:
            serve_script("deploy.py", args=["--verbose", "--dry-run"])
            serve_script("cli.py", dynamic=True)  # Args from URL: /cli?args=value1,value2
            serve_script("cli.py", dynamic=True, args_param="with")  # Args from URL: /cli?with=value1,value2
    """
    kwargs = {}
    if port != 8000:
        kwargs["port"] = port
    if title is not None:
        kwargs["title"] = title
    if theme is not None:
        kwargs["theme"] = theme
    if log_level != "info":
        kwargs["log_level"] = log_level
    if args is not None:
        kwargs["args"] = args
    if dynamic:
        kwargs["dynamic"] = dynamic
    if args_param != "args":
        kwargs["args_param"] = args_param

    cfg = _prepare_config(None, True, **kwargs)
    cfg._target = Path(script_path)
    cfg._mode = "script"

    _auto_generate_title(cfg, "script", cfg._target, kwargs)
    ServeWithConfig.serve(cfg)


def serve_function(
    func: Callable,
    port: int = 8000,
    title: Optional[str] = None,
    theme: Optional[Dict[str, str]] = None,
    log_level: Optional[str] = "info",
    args: Optional[List[str]] = None,
    dynamic: bool = False,
    args_param: str = "args",
) -> None:
    """Serve a Python function in a browser terminal.

    Creates a web-accessible terminal that runs the provided Python function.

    Args:
        func: The function to serve in the terminal
        port: Web server port (default: 8000)
        title: Terminal window title (default: auto-generated from function name)
        theme: Terminal theme colors (default: {"background": "black", "foreground": "white"})
        log_level: Logging level ("debug", "info", "warning", "error", None) (default: "info")
        args: Command-line arguments to pass to the function via sys.argv (default: None)
        dynamic: Enable dynamic arguments via URL query parameters (default: False)
        args_param: Name of the query parameter for dynamic arguments (default: "args")

    Examples:
        Basic usage:
            serve_function(my_function)

        With custom configuration:
            serve_function(my_function, port=8080, title="My CLI Tool")
            serve_function(my_function, theme={"background": "navy", "foreground": "white"})
            serve_function(my_function, log_level="debug")
        
        With arguments (compatible with Click, Fire, argparse):
            serve_function(my_cli_function, args=["--verbose", "--output", "result.txt"])
            serve_function(my_cli_function, dynamic=True)  # Args from URL: /func?args=verbose,output
            serve_function(my_cli_function, dynamic=True, args_param="with")  # Args from URL: /func?with=verbose,output

    Note:
        For advanced configuration options like environment variables, authentication,
        or custom templates, use serve_apps() instead.
    """
    kwargs = {}
    if port != 8000:
        kwargs["port"] = port
    if title is not None:
        kwargs["title"] = title
    if theme is not None:
        kwargs["theme"] = theme
    if log_level != "info":
        kwargs["log_level"] = log_level
    if args is not None:
        kwargs["args"] = args
    if dynamic:
        kwargs["dynamic"] = dynamic
    if args_param != "args":
        kwargs["args_param"] = args_param

    cfg = _prepare_config(None, True, **kwargs)
    cfg._target = func
    cfg._mode = "function"

    _auto_generate_title(cfg, "function", func, kwargs)
    ServeWithConfig.serve(cfg)


def serve_apps(
    app: FastAPI,
    terminal_routes: Dict[
        str, Union[str, Path, List, Dict[str, Any], Callable, AutoIndex]
    ],
    config: Optional[TerminaideConfig] = None,
    banner: Union[bool, str] = True,
    log_level: Optional[str] = "info",
    **kwargs,
) -> None:
    """Integrate multiple terminals and index pages into a FastAPI application.

    Configures a FastAPI application to serve multiple terminal instances and/or
    index pages at different routes.

    Args:
        app: FastAPI application to extend
        terminal_routes: Dictionary mapping paths to scripts, functions, or index pages
        config: Configuration options for the terminals
        banner: Controls banner display (default: True)
        log_level: Logging level ("debug", "info", "warning", "error", None) (default: "info")
        **kwargs: Additional configuration overrides

    Terminal Routes Configuration:
        Each value in terminal_routes can be:
        - String/Path: Script file path
        - Callable: Python function
        - AutoIndex: Navigable menu page (HTML or Curses)
        - List: [script_path, arg1, arg2, ...] for scripts with arguments
        - Dict: Advanced configuration with "script"/"function" key plus options

    Common Configuration Options:
        - port: Web server port (default: 8000)
        - title: Terminal window title (default: auto-generated)
        - theme: Terminal theme colors
        - ttyd_port: Base port for ttyd processes (default: 7681)
        - mount_path: Base path for terminal mounting (default: "/")
        - preview_image: Default preview image for social media sharing

    Examples:
        Simple terminal routes:
            serve_apps(app, {
                "/script": "my_script.py",
                "/hello": my_function,
                "/": AutoIndex(type="html", title="MENU", menu=[...])
            })

        Advanced configuration:
            serve_apps(app, {
                "/deploy": ["deploy.py", "--verbose"],
                "/admin": {
                    "function": admin_function,
                    "title": "Admin Terminal",
                    "preview_image": "admin.png"
                },
                "/cli": {
                    "script": "cli.py",
                    "dynamic": True,
                    "args_param": "with"  # Use ?with=arg1,arg2 instead of ?args=arg1,arg2
                }
            }, log_level="debug")

    Note:
        For simple single-terminal applications, consider using serve_function
        or serve_script instead.
    """
    if not terminal_routes:
        logger.warning(
            "No terminal routes provided to serve_apps(). No terminals will be served."
        )
        return

    if log_level != "info":
        kwargs["log_level"] = log_level
    
    cfg = _prepare_config(config, banner, **kwargs)
    cfg._target = terminal_routes
    cfg._app = app
    cfg._mode = "apps"

    ServeWithConfig.serve(cfg)


################################################################################
# UI Components & Utilities
################################################################################

# UI Components are imported and re-exported
# AutoIndex - Create navigable index pages (HTML or Curses)

# Utilities are imported and re-exported
# terminascii - Generate ASCII banners
# ServerMonitor - Process output monitoring with rich terminal interface

################################################################################
# Public API Exports
################################################################################

__all__ = [
    # Solo Server API
    "serve_function",
    "serve_script",
    # Apps Server API
    "serve_apps",
    # UI Components
    "AutoIndex",
    # Utilities
    "terminascii",
    "ServerMonitor",
]


# server.py

"""
Application runner implementation for Terminaide.

This module contains the ServeWithConfig class that handles running applications
in different Terminaide modes (function, script, apps).
"""

import os
import sys
import signal
import logging
import uvicorn
from pathlib import Path
from fastapi import FastAPI
from contextlib import asynccontextmanager
from starlette.middleware.base import BaseHTTPMiddleware

from .config import (
    TerminaideConfig,
    convert_terminaide_config_to_ttyd_config,
    terminaide_lifespan,
    smart_resolve_path,
)
from .wrappers import (
    generate_function_wrapper,
    get_or_ensure_function_wrapper,
)

logger = logging.getLogger("terminaide")


class _ProxyHeaderMiddleware(BaseHTTPMiddleware):
    """
    Middleware that detects and respects common proxy headers for HTTPS, enabling
    terminaide to work correctly behind load balancers and proxies.
    """

    async def dispatch(self, request, call_next):
        # Check X-Forwarded-Proto (most common)
        forwarded_proto = request.headers.get("x-forwarded-proto")
        if forwarded_proto == "https":
            original_scheme = request.scope.get("scheme", "unknown")
            request.scope["scheme"] = "https"

            # Log this detection once per deployment to help with debugging
            logger.debug(
                f"HTTPS detected via X-Forwarded-Proto header "
                f"(original scheme: {original_scheme})"
            )

        # Check Forwarded header (RFC 7239)
        forwarded = request.headers.get("forwarded")
        if forwarded and "proto=https" in forwarded.lower():
            request.scope["scheme"] = "https"

        # AWS Elastic Load Balancer sometimes uses this
        elb_proto = request.headers.get("x-forwarded-protocol")
        if elb_proto == "https":
            request.scope["scheme"] = "https"

        return await call_next(request)


class ServeWithConfig:
    """Class responsible for handling the serving implementation for different modes."""

    @staticmethod
    def add_proxy_middleware_if_needed(app: FastAPI, config: TerminaideConfig) -> None:
        """
        Adds proxy header middleware if trust_proxy_headers=True in config.
        This ensures that X-Forwarded-Proto from proxies like ngrok is respected,
        preventing mixed-content errors behind HTTPS tunnels or load balancers.
        """

        if config.trust_proxy_headers:
            try:
                if not any(
                    m.cls.__name__ == "_ProxyHeaderMiddleware"
                    for m in getattr(app, "user_middleware", [])
                ):
                    app.add_middleware(_ProxyHeaderMiddleware)
                    logger.info("Added proxy header middleware for HTTPS detection")

            except Exception as e:
                logger.warning(f"Failed to add middleware: {e}")

    @classmethod
    def display_banner(cls, mode, banner_value):
        """Display a banner based on the banner parameter value.

        Args:
            mode: The serving mode (function, script, apps)
            banner_value: True for Rich panel, False for no banner, or a string to print directly
        """
        if os.environ.get("TERMINAIDE_BANNER_SHOWN") == "1":
            return
        os.environ["TERMINAIDE_BANNER_SHOWN"] = "1"

        # Handle string banner - print it directly
        if isinstance(banner_value, str):
            print(banner_value)
            logger.debug(f"Starting Terminaide in {mode.upper()} mode")
            return

        # Handle boolean False - no banner
        if banner_value is False:
            logger.debug(
                f"Starting Terminaide in {mode.upper()} mode (banner disabled)"
            )
            return

        # Handle boolean True - show Rich panel
        try:
            from rich.console import Console
            from rich.panel import Panel
            from rich.align import Align

            mode_colors = {
                "function": "dark_orange",
                "script": "blue",
                "apps": "magenta",
            }
            color = mode_colors.get(mode, "yellow")
            mode_upper = mode.upper()
            console = Console(highlight=False)

            # Center the content within the panel
            centered_content = Align.center(f"TERMINAIDE {mode_upper} SERVER")
            panel = Panel(
                centered_content,
                border_style=color,
                expand=True,
                padding=(0, 1),
            )
            console.print(panel)
        except ImportError:
            mode_upper = mode.upper()
            banner = f"== TERMINAIDE SERVING IN {mode_upper} MODE =="
            print(f"\033[1m\033[92m{banner}\033[0m")

        logger.debug(f"Starting Terminaide in {mode.upper()} mode")

    @classmethod
    def serve(cls, config) -> None:
        """Serves the application based on the configuration mode."""
        # Configure logging based on config
        from .logger import setup_package_logging

        setup_package_logging(log_level=config.log_level)
        
        logger.debug(f"Logging configured with log_level={config.log_level}, level={logger.level}")

        # Display banner based on config.banner value
        if config.banner:
            cls.display_banner(config._mode, config.banner)

        if config._mode == "function":
            cls.serve_function(config)
        elif config._mode == "script":
            cls.serve_script(config)
        elif config._mode == "apps":
            cls.serve_apps(config)
        else:
            raise ValueError(f"Unknown serving mode: {config._mode}")

    @classmethod
    def serve_function(cls, config) -> None:
        """Implementation for serving a function."""
        # Direct mode - use lazy validation for reliable wrapper access
        func = config._target
        ephemeral_path = get_or_ensure_function_wrapper(func, args=config.args, config=config)

        # Import from factory to avoid circular dependency
        from .factory import copy_config_attributes

        script_config = copy_config_attributes(config)
        script_config._target = ephemeral_path
        script_config._mode = "function"
        script_config._original_function_name = func.__name__

        logger.debug(f"Using title: {script_config.title} for function {func.__name__}")

        cls.serve_script(script_config)

    @classmethod
    def serve_script(cls, config) -> None:
        """Implementation for serving a script."""
        script_path = config._target
        if not isinstance(script_path, Path):
            script_path = Path(script_path)

        script_path = smart_resolve_path(script_path)
        if not script_path.exists():
            print(f"\033[91mError: Script not found: {script_path}\033[0m")
            return

        # Direct mode
        ttyd_config = convert_terminaide_config_to_ttyd_config(config, script_path)
        # Import from factory to avoid circular dependency
        from .factory import create_app_with_lifespan

        app = create_app_with_lifespan(config.title, config, ttyd_config)

        def handle_exit(sig, _):
            print("\033[93mShutting down...\033[0m")
            
            # Clean up ephemeral files on graceful shutdown
            try:
                from .wrappers import cleanup_own_ephemeral_files
                cleanup_own_ephemeral_files()
            except ImportError:
                pass  # Graceful fallback if import fails
            
            sys.exit(0)

        signal.signal(signal.SIGINT, handle_exit)
        signal.signal(signal.SIGTERM, handle_exit)

        uvicorn.run(
            app,
            host="0.0.0.0",
            port=config.port,
            log_level=config.log_level or "info",
        )

    @classmethod
    def serve_apps(cls, config) -> None:
        """Implementation for serving multiple apps."""
        # Display banner if enabled, for consistency with other serve methods
        if config.banner:
            cls.display_banner(config._mode, config.banner)

        logger.debug(f"serve_apps called with log_level={config.log_level}")
        
        app = config._app
        terminal_routes = config._target

        # Process function-based routes to generate ephemeral script wrappers
        ttyd_config = convert_terminaide_config_to_ttyd_config(config)

        # Generate wrapper scripts for all function-based routes
        for script_config in ttyd_config.script_configs:
            if script_config.is_function_based:
                func = script_config.function_object
                if func is not None:
                    logger.debug(
                        f"Generating wrapper script for function '{func.__name__}' at route {script_config.route_path}"
                    )
                    wrapper_path = get_or_ensure_function_wrapper(func, args=script_config.args, config=config)
                    script_config.set_function_wrapper_path(wrapper_path)
                    logger.debug(
                        f"Function '{func.__name__}' will use wrapper script at {wrapper_path}"
                    )

        # Add middleware silently, we'll log during startup
        if config.trust_proxy_headers:
            try:
                if not any(
                    m.cls.__name__ == "_ProxyHeaderMiddleware"
                    for m in getattr(app, "user_middleware", [])
                ):
                    app.add_middleware(_ProxyHeaderMiddleware)
                    # Store the flag for logging during lifespan startup
                    app.state.terminaide_middleware_added = True

            except Exception as e:
                logger.warning(f"Failed to add proxy header middleware: {e}")

        # Rest of the method remains the same...
        sentinel_attr = "_terminaide_lifespan_attached"
        if getattr(app.state, sentinel_attr, False):
            return

        setattr(app.state, sentinel_attr, True)

        original_lifespan = app.router.lifespan_context

        @asynccontextmanager
        async def terminaide_merged_lifespan(_app: FastAPI):
            # Log middleware addition at startup (after banner has been shown)
            if getattr(_app.state, "terminaide_middleware_added", False):
                logger.info("Added proxy header middleware for HTTPS detection")
                # Clear the flag so we don't log again
                delattr(_app.state, "terminaide_middleware_added")

            if original_lifespan is not None:
                async with original_lifespan(_app):
                    async with terminaide_lifespan(_app, ttyd_config):
                        yield
            else:
                async with terminaide_lifespan(_app, ttyd_config):
                    yield

        app.router.lifespan_context = terminaide_merged_lifespan



# index.py

"""Unified index functionality for HTML and terminal interfaces."""

import curses
import signal
import sys
import logging
import subprocess
import importlib
from typing import Optional, List, Dict, Any, Union, Literal
from pathlib import Path
from functools import lru_cache

from .terminascii import terminascii

logger = logging.getLogger("terminaide")

# Global state for signal handling (curses mode)
stdscr = None
exit_requested = False


# Base Classes


class BaseMenuItem:
    """Base menu item with path, title, and optional execution parameters."""

    __slots__ = ("path", "title", "function", "script", "launcher_args", "new_tab")

    def __init__(
        self, path: str, title: str, function=None, script=None, launcher_args=None, new_tab=None
    ):
        self.path = path
        self.title = title
        self.function = function
        self.script = script
        self.launcher_args = launcher_args or {}
        self.new_tab = new_tab

    def is_external(self) -> bool:
        """Check if this is an external URL."""
        return self.path.startswith(("http://", "https://"))

    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary for JSON serialization."""
        result = {"path": self.path, "title": self.title}
        if self.new_tab is not None:
            result["new_tab"] = self.new_tab
        return result


class BaseIndex:
    """Base configuration for index/menu pages."""

    __slots__ = (
        "menu_items",
        "subtitle",
        "epititle",
        "title",
        "supertitle",
        "preview_image",
        "instructions",
        "_all_items_cache",
    )

    def __init__(
        self,
        menu: Union[List[Dict[str, Any]], Dict[str, Any]],
        subtitle: Optional[str] = None,
        epititle: Optional[str] = None,
        title: Optional[str] = None,
        supertitle: Optional[str] = None,
        preview_image: Optional[Union[str, Path]] = None,
        instructions: Optional[str] = None,
    ):
        self._all_items_cache = None
        self._parse_and_validate_menu(menu)
        self.subtitle = subtitle
        self.epititle = epititle
        self.title = title or "Index"
        self.supertitle = supertitle
        self.preview_image = Path(preview_image) if preview_image else None
        self.instructions = instructions

    def _parse_and_validate_menu(self, menu):
        """Parse and validate the menu structure."""
        if not isinstance(menu, list):
            raise ValueError("Menu must be a list of menu items")
        
        if not menu:
            raise ValueError("Menu must contain at least one item")
        
        for i, item in enumerate(menu):
            if not isinstance(item, dict):
                raise ValueError(f"Menu item at index {i} must be a dictionary")
            if "title" not in item:
                raise ValueError(f"Menu item at index {i} missing required 'title' key")
            
            # Must have one of: path, function, or script
            if not any(key in item for key in ["path", "function", "script"]):
                raise ValueError(f"Menu item at index {i} must have 'path', 'function', or 'script'")
        
        # Create menu items directly
        self.menu_items = [self._create_menu_item(item) for item in menu]

    def _create_menu_item(self, item):
        """Create a menu item from dict. Override in subclasses for specialized items."""
        if isinstance(item, BaseMenuItem):
            return item
        return BaseMenuItem(
            path=item.get("path", ""),
            title=item.get("title", ""),
            function=item.get("function"),
            script=item.get("script"),
            launcher_args=item.get("launcher_args", {}),
            new_tab=item.get("new_tab"),
        )

    def get_all_menu_items(self) -> List[BaseMenuItem]:
        """Get all menu items as a flat list."""
        if self._all_items_cache is None:
            self._all_items_cache = self.menu_items
        return self._all_items_cache

    def __repr__(self) -> str:
        """String representation for debugging."""
        item_count = len(self.get_all_menu_items())
        class_name = self.__class__.__name__
        return f"{class_name}(title='{self.title}', items={item_count})"


# Auto Classes


class AutoMenuItem(BaseMenuItem):
    """Extended BaseMenuItem that can handle both HTML and Curses behavior."""

    __slots__ = ()

    def launch(self) -> bool:
        """Launch this menu item (Curses mode only)."""
        return launch_menu_item(self)


class AutoIndex(BaseIndex):
    """Unified configuration for index/menu pages with automatic type selection."""

    __slots__ = ("index_type", "_template_context_cache")

    def __init__(
        self,
        type: Literal["html", "curses"],
        menu: Union[List[Dict[str, Any]], Dict[str, Any]],
        subtitle: Optional[str] = None,
        epititle: Optional[str] = None,
        title: Optional[str] = None,
        supertitle: Optional[str] = None,
        preview_image: Optional[Union[str, Path]] = None,
        instructions: Optional[str] = None,
    ):
        if type not in ("html", "curses"):  # Use tuple for faster membership test
            raise ValueError(f"type must be 'html' or 'curses', got: {type}")

        self.index_type = type
        self._template_context_cache = None
        super().__init__(
            menu=menu,
            subtitle=subtitle,
            epititle=epititle,
            title=title,
            supertitle=supertitle,
            preview_image=preview_image,
            instructions=instructions,
        )


    def _create_menu_item(self, item):
        """Create an AutoMenuItem from dict or existing item."""
        if isinstance(item, AutoMenuItem):
            return item
        return AutoMenuItem(
            path=item.get("path", ""),
            title=item.get("title", ""),
            function=item.get("function"),
            script=item.get("script"),
            launcher_args=item.get("launcher_args", {}),
            new_tab=item.get("new_tab"),
        )

    def get_all_menu_items(self) -> List[AutoMenuItem]:
        """Get all menu items as a flat list."""
        if self._all_items_cache is None:
            self._all_items_cache = self.menu_items
        return self._all_items_cache

    def to_template_context(self) -> Dict[str, Any]:
        """Convert to dictionary for template rendering (HTML only)."""
        if self._template_context_cache is None:
            self._template_context_cache = get_template_context(self)
        return self._template_context_cache

    def show(self) -> Optional[str]:
        """Display the curses menu and launch selected items (Curses only)."""
        if self.index_type != "curses":
            raise AttributeError("show() is only available for Curses indexes")
        return show_curses_menu(self)

    def __repr__(self) -> str:
        """String representation for debugging."""
        item_count = len(self.get_all_menu_items())
        return f"AutoIndex(type='{self.index_type}', title='{self.title}', items={item_count})"


# Curses Implementation


def handle_exit(*_):
    """Handle SIGINT (Ctrl+C) for clean program exit."""
    global exit_requested
    exit_requested = True


def cleanup():
    """Restore terminal state and print goodbye message."""
    global stdscr
    try:
        if stdscr:
            curses.endwin()
            stdscr = None
        print("\033[?25h\033[2J\033[H", end="")  # Show cursor, clear screen
        print("Thank you for using terminaide")
        print("Goodbye!")
        sys.stdout.flush()
    except Exception:
        pass


def safe_addstr(win, y, x, text, attr=0):
    """Safely add a string to the screen, handling boundary conditions."""
    h, w = win.getmaxyx()
    if y < 0 or y >= h or x < 0 or x >= w:
        return
    text = text[: w - x] if w - x > 0 else ""
    try:
        win.addstr(y, x, text, attr)
    except curses.error:
        pass


def launch_menu_item(menu_item: AutoMenuItem) -> bool:
    """Launch a menu item (Curses mode only). Always returns True to return to menu."""
    try:
        if menu_item.function:
            menu_item.function()
        elif menu_item.script:
            subprocess.run(
                [sys.executable, menu_item.script], capture_output=False, check=False
            )
        elif menu_item.path and not menu_item.is_external():
            return _launch_from_path(menu_item.path)
        else:
            logger.warning(f"No launch method defined for {menu_item.title}")
    except Exception as e:
        target = getattr(
            menu_item.function, "__name__", menu_item.script or menu_item.path
        )
        logger.error(f"Error launching {target}: {e}")
    return True


# Pre-compiled set for faster membership testing
_GAME_PATHS = frozenset(["snake", "tetris", "pong", "asteroids"])


def _launch_from_path(path: str) -> bool:
    """Dynamically resolve and launch a function or module from a path string."""
    try:
        if "." in path:
            parts = path.split(".")
            function_name = parts[-1]
            module_path = ".".join(parts[:-1])

            module = sys.modules.get(module_path)
            if module is not None:
                module = importlib.reload(module)
            else:
                module = importlib.import_module(module_path)

            # Reset module-level state if present (for games)
            for attr in ("exit_requested", "stdscr"):
                if hasattr(module, attr):
                    setattr(module, attr, False if attr == "exit_requested" else None)

            if hasattr(module, function_name):
                getattr(module, function_name)()
                return True
            else:
                logger.error(
                    f"Function {function_name} not found in module {module_path}"
                )
                return True
        else:
            # Single name - try common patterns for backwards compatibility with terminarcade games
            if path in _GAME_PATHS:
                try:
                    module_path = f"terminarcade.{path}"
                    importlib.import_module(module_path)
                except ImportError:
                    module_path = f"terminaide.terminarcade.{path}"
                function_name = f"play_{path}"

                module = sys.modules.get(module_path)
                if module is not None:
                    module = importlib.reload(module)
                else:
                    module = importlib.import_module(module_path)

                # Reset module-level state
                for attr in ("exit_requested", "stdscr"):
                    if hasattr(module, attr):
                        setattr(
                            module, attr, False if attr == "exit_requested" else None
                        )

                getattr(module, function_name)()
                return True
            else:
                logger.warning(f"Don't know how to launch path: {path}")
                return True
    except Exception as e:
        logger.error(f"Error launching from path {path}: {e}")
        return True


def show_curses_menu(auto_index: AutoIndex) -> Optional[str]:
    """Display the curses menu and launch selected items."""
    global exit_requested
    last_selection = None

    signal.signal(signal.SIGINT, handle_exit)

    try:
        while True:
            choice = curses.wrapper(lambda stdscr: _index_menu_loop(stdscr, auto_index))

            if choice == "exit" or exit_requested or choice is None:
                return last_selection

            if hasattr(choice, "launch"):
                last_selection = choice.path
                launch_menu_item(choice)
            else:
                return choice

    except Exception as e:
        logger.error(f"Error in AutoIndex: {e}")
        return last_selection
    finally:
        exit_requested = True
        cleanup()


def _index_menu_loop(stdscr_param, auto_index: AutoIndex):
    """Main menu interface for curses mode."""
    global stdscr, exit_requested
    stdscr = stdscr_param
    exit_requested = False

    signal.signal(signal.SIGINT, handle_exit)

    # Configure terminal
    curses.start_color()
    curses.use_default_colors()
    curses.init_pair(1, curses.COLOR_BLUE, -1)  # Title
    curses.init_pair(2, curses.COLOR_WHITE, -1)  # Instructions & Subtitle
    curses.init_pair(3, curses.COLOR_CYAN, -1)  # Supertitle
    curses.init_pair(4, curses.COLOR_BLACK, curses.COLOR_CYAN)  # Unselected
    curses.init_pair(5, curses.COLOR_BLACK, curses.COLOR_WHITE)  # Selected
    curses.init_pair(6, curses.COLOR_GREEN, -1)  # Menu labels
    curses.init_pair(7, curses.COLOR_WHITE, -1)  # Epititle
    curses.init_pair(8, curses.COLOR_WHITE, -1)  # Epititle dim

    curses.curs_set(0)  # Hide cursor
    stdscr.clear()

    # Current state
    current_option = 0
    previous_option = 0

    # Get screen dimensions
    my, mx = stdscr.getmaxyx()

    # Generate ASCII art from title
    if auto_index.title:
        ascii_art = terminascii(auto_index.title)
        if ascii_art:
            title_lines = ascii_art.split("\n")
            # Remove empty trailing lines in one pass
            while title_lines and not title_lines[-1].strip():
                title_lines.pop()
        else:
            title_lines = [auto_index.title]
    else:
        title_lines = ["Index"]

    # Calculate layout positions
    current_y = 1

    # Draw supertitle if provided
    if auto_index.supertitle:
        safe_addstr(
            stdscr,
            current_y,
            (mx - len(auto_index.supertitle)) // 2,
            auto_index.supertitle,
            curses.color_pair(3) | curses.A_BOLD,
        )
        current_y += 2

    # Draw title
    for i, line in enumerate(title_lines):
        if len(line) <= mx:
            safe_addstr(
                stdscr,
                current_y + i,
                (mx - len(line)) // 2,
                line,
                curses.color_pair(1) | curses.A_BOLD,
            )
    current_y += len(title_lines) + 1

    # Draw subtitle if provided
    if auto_index.subtitle:
        safe_addstr(
            stdscr,
            current_y,
            (mx - len(auto_index.subtitle)) // 2,
            auto_index.subtitle,
            curses.color_pair(2),
        )
        current_y += 2

    # Store menu start position
    menu_start_y = current_y

    def draw_menu():
        """Draw the menu."""
        nonlocal current_y
        y_pos = menu_start_y

        # Clear menu area
        for clear_y in range(menu_start_y, my - 3):
            safe_addstr(stdscr, clear_y, 0, " " * mx)

        # Draw instructions if provided
        if auto_index.instructions:
            safe_addstr(
                stdscr,
                y_pos,
                (mx - len(auto_index.instructions)) // 2,
                auto_index.instructions,
                curses.color_pair(6) | curses.A_BOLD,
            )
            y_pos += 2

        # Calculate button layout
        menu_items = auto_index.get_all_menu_items()
        options = [item.title for item in menu_items]
        button_padding = 4
        button_width = max(len(o) for o in options) + 6
        num_options = len(options)
        total_buttons_width = (button_width * num_options) + (
            button_padding * (num_options - 1)
        )

        # Center the row of buttons
        start_x = max(0, (mx - total_buttons_width) // 2)
        menu_y = y_pos

        # Draw menu options horizontally
        for i, option in enumerate(options):
            button_x = start_x + (i * (button_width + button_padding))
            if button_x + button_width > mx:
                break  # Skip if button would go off screen

            st = curses.color_pair(5) if i == current_option else curses.color_pair(4)

            # Center the text within the button
            text_padding = (button_width - len(option)) // 2
            button_text = (
                " " * text_padding
                + option
                + " " * (button_width - len(option) - text_padding)
            )

            safe_addstr(stdscr, menu_y, button_x, button_text, st | curses.A_BOLD)

        return menu_y + 2

    # Initial menu draw
    draw_menu()

    # Draw epititle at bottom if provided
    if auto_index.epititle:
        epititle_lines = auto_index.epititle.split("\n")
        total_lines = len(epititle_lines)
        start_y = my - total_lines - 1

        for i, line in enumerate(epititle_lines):
            y_pos = start_y + i
            x_pos = (mx - len(line)) // 2
            safe_addstr(
                stdscr,
                y_pos,
                x_pos,
                line,
                curses.color_pair(8) | curses.A_DIM | curses.A_ITALIC,
            )

    # Main menu loop
    while True:
        if exit_requested:
            break

        # Update menu selection if changed
        if current_option != previous_option:
            draw_menu()
            previous_option = current_option

        stdscr.refresh()

        try:
            k = stdscr.getch()

            if k in [ord("q"), ord("Q"), 27]:  # q, Q, or ESC
                break
            elif k in [curses.KEY_LEFT, ord("a"), ord("A")] and current_option > 0:
                current_option -= 1
            elif (
                k in [curses.KEY_RIGHT, ord("d"), ord("D")]
                and current_option < len(auto_index.get_all_menu_items()) - 1
            ):
                current_option += 1
            elif k in [curses.KEY_ENTER, ord("\n"), ord("\r")]:
                selected_item = auto_index.get_all_menu_items()[current_option]
                return selected_item
        except KeyboardInterrupt:
            break

    return "exit"


# HTML Implementation


@lru_cache(maxsize=128)
def is_ascii_art(text: str) -> bool:
    """Detect if a string contains ASCII art by checking for multiple lines."""
    return "\n" in text and len(text.split("\n")) > 1


def get_template_context(auto_index: AutoIndex) -> Dict[str, Any]:
    """Convert AutoIndex to dictionary for template rendering (HTML only)."""
    if auto_index.index_type != "html":
        raise AttributeError(
            "get_template_context() is only available for HTML indexes"
        )

    # Generate ASCII art for title
    title_ascii = None
    if auto_index.title and auto_index.title != "Index":
        title_ascii = (
            auto_index.title
            if is_ascii_art(auto_index.title)
            else terminascii(auto_index.title)
        )

    # Prepare menu items data for JavaScript
    menu_items = [item.to_dict() for item in auto_index.get_all_menu_items()]
    total_items = len(menu_items)

    return {
        "page_title": auto_index.title,
        "title_ascii": title_ascii,
        "supertitle": auto_index.supertitle,
        "subtitle": auto_index.subtitle,
        "epititle": auto_index.epititle,
        "instructions": auto_index.instructions,
        "menu_items": menu_items,
        "total_items": total_items,
        "title": auto_index.title,
    }



# snake.py

import curses
import random
import signal
import sys
from collections import deque

stdscr = None
exit_requested = False


def _snake_game_loop(stdscr_param):
    """Main snake game function that handles the game loop.

    Args:
        stdscr_param: The curses window.
    """
    global stdscr, exit_requested
    stdscr = stdscr_param
    exit_requested = False

    signal.signal(signal.SIGINT, handle_exit)
    setup_terminal(stdscr)
    max_y, max_x = stdscr.getmaxyx()
    ph, pw = max_y - 2, max_x - 2
    high_score = 0

    while True:
        if exit_requested:
            cleanup()
            return

        score = run_game(stdscr, max_y, max_x, ph, pw, high_score)

        if exit_requested:
            cleanup()
            return

        high_score = max(high_score, score)

        if show_game_over(stdscr, score, high_score, max_y, max_x):
            break


def setup_terminal(stdscr):
    """Configure terminal settings for the game."""
    curses.curs_set(0)
    curses.noecho()
    curses.cbreak()
    stdscr.keypad(True)
    curses.use_env(False)
    curses.start_color()
    curses.use_default_colors()
    curses.init_pair(1, curses.COLOR_GREEN, -1)
    curses.init_pair(2, curses.COLOR_CYAN, -1)
    curses.init_pair(3, curses.COLOR_RED, -1)
    curses.init_pair(4, curses.COLOR_WHITE, -1)
    curses.init_pair(5, curses.COLOR_YELLOW, -1)


def run_game(stdscr, my, mx, ph, pw, high_score=0):
    """Run the main game loop.

    Args:
        stdscr: The curses window.
        my, mx: Maximum y and x coordinates.
        ph, pw: Playable height and width.
        high_score: Current high score.

    Returns:
        int: The final score.
    """
    global exit_requested

    score = 0
    speed = 100
    win = curses.newwin(ph + 2, pw + 2, 0, 0)
    win.keypad(True)
    win.timeout(speed)

    s = deque([(ph // 2, pw // 4)])
    direction = curses.KEY_RIGHT
    food = new_food(s, ph, pw)

    draw_screen(stdscr, win, s, food, score, high_score, mx)

    while True:
        if exit_requested:
            cleanup()
            return score

        key = win.getch()

        # Check for exit key
        if key in (ord("q"), 27):  # q or ESC
            cleanup()
            return score

        new_dir = process_input(key, direction)
        if new_dir:
            direction = new_dir

        hy, hx = s[0]
        nh = move_head(hy, hx, direction)

        if is_collision(nh, s, ph, pw):
            break

        s.appendleft(nh)

        if nh == food:
            score += 10
            if speed > 50:
                speed = max(50, speed - 3)
                win.timeout(speed)
            food = new_food(s, ph, pw)
        else:
            s.pop()

        draw_screen(stdscr, win, s, food, score, high_score, mx)

    return score


def draw_screen(stdscr, win, snake, food, score, high_score, mx):
    """Draw the game screen with all elements."""
    win.erase()
    draw_border(win)
    try:
        win.addch(
            food[0] + 1, food[1] + 1, ord("*"), curses.color_pair(3) | curses.A_BOLD
        )
    except:
        curses.error
    draw_snake(win, snake)
    draw_score(stdscr, score, high_score, mx)
    stdscr.noutrefresh()
    win.noutrefresh()
    curses.doupdate()


def draw_border(win):
    """Draw the game border with title."""
    win.box()
    title = "PLAY SNAKE"
    w = win.getmaxyx()[1]
    if w > len(title) + 4:
        win.addstr(
            0, (w - len(title)) // 2, title, curses.A_BOLD | curses.color_pair(5)
        )


def draw_snake(win, snake):
    """Draw the snake on the game window."""
    try:
        y, x = snake[0]
        win.addch(y + 1, x + 1, ord("O"), curses.color_pair(1) | curses.A_BOLD)
        for y, x in list(snake)[1:]:
            win.addch(y + 1, x + 1, ord("o"), curses.color_pair(2))
    except:
        curses.error


def draw_score(stdscr, score, high_score, mx):
    """Draw the score display at the top of the screen."""
    stdscr.addstr(0, 0, " " * mx)
    stdscr.addstr(0, 2, f" Score: {score} ", curses.color_pair(5) | curses.A_BOLD)
    txt = f" High Score: {high_score} "
    stdscr.addstr(0, mx - len(txt) - 2, txt, curses.color_pair(5) | curses.A_BOLD)


def process_input(key, cur_dir):
    """Process keyboard input for snake movement.

    Args:
        key: The key pressed.
        cur_dir: Current direction.

    Returns:
        The new direction or None if no valid key was pressed.
    """
    if key in [curses.KEY_UP, ord("w"), ord("W")] and cur_dir != curses.KEY_DOWN:
        return curses.KEY_UP
    if key in [curses.KEY_DOWN, ord("s"), ord("S")] and cur_dir != curses.KEY_UP:
        return curses.KEY_DOWN
    if key in [curses.KEY_LEFT, ord("a"), ord("A")] and cur_dir != curses.KEY_RIGHT:
        return curses.KEY_LEFT
    if key in [curses.KEY_RIGHT, ord("d"), ord("D")] and cur_dir != curses.KEY_LEFT:
        return curses.KEY_RIGHT
    return None


def move_head(y, x, d):
    """Calculate new head position based on direction.

    Args:
        y, x: Current head position.
        d: Direction to move.

    Returns:
        tuple: New head position (y, x).
    """
    if d == curses.KEY_UP:
        return (y - 1, x)
    if d == curses.KEY_DOWN:
        return (y + 1, x)
    if d == curses.KEY_LEFT:
        return (y, x - 1)
    return (y, x + 1)


def is_collision(head, snake, h, w):
    """Check if the snake has collided with wall or itself.

    Args:
        head: Snake head position.
        snake: Snake body positions.
        h, w: Game board height and width.

    Returns:
        bool: True if collision detected, False otherwise.
    """
    y, x = head
    if y < 0 or y >= h or x < 0 or x >= w:
        return True
    if head in list(snake)[1:]:
        return True
    return False


def new_food(snake, h, w):
    """Generate new food position.

    Args:
        snake: Current snake body positions.
        h, w: Game board height and width.

    Returns:
        tuple: New food position (y, x).
    """
    while True:
        fy = random.randint(0, h - 1)
        fx = random.randint(0, w - 1)
        if (fy, fx) not in snake:
            return (fy, fx)


def show_game_over(stdscr, score, high_score, my, mx):
    """Show game over screen and handle restart/quit options.

    Returns:
        bool: True if user chooses to quit, False to restart.
    """
    stdscr.clear()
    cy = my // 2
    data = [
        ("GAME OVER", -3, curses.A_BOLD | curses.color_pair(3)),
        (f"Your Score: {score}", -1, curses.color_pair(5)),
        (f"High Score: {high_score}", 0, curses.color_pair(5)),
        ("Press 'r' to restart", 2, 0),
        ("Press 'q' to quit", 3, 0),
    ]
    for txt, yo, attr in data:
        stdscr.addstr(cy + yo, mx // 2 - len(txt) // 2, txt, attr)
    stdscr.noutrefresh()
    curses.doupdate()
    stdscr.nodelay(False)
    while True:
        k = stdscr.getch()
        if k == ord("q"):
            return True
        if k == ord("r"):
            return False


def cleanup():
    """Clean up terminal state when exiting."""
    if stdscr:
        try:
            curses.endwin()
            print("\033[?25l\033[2J\033[H", end="")
            try:
                rows, cols = stdscr.getmaxyx()
            except:
                rows, cols = 24, 80
            msg = "Thanks for playing Snake!"
            print("\033[2;{}H{}".format((cols - len(msg)) // 2, msg))
            print("\033[3;{}H{}".format((cols - len("Goodbye!")) // 2, "Goodbye!"))
            sys.stdout.flush()
        except:
            pass


def handle_exit(sig, frame):
    """Handle SIGINT (Ctrl+C) for program exit."""
    global exit_requested
    exit_requested = True


def snake():
    """Run the snake game from command line.

    This is the main public-facing function for launching the snake game.
    """
    try:
        curses.wrapper(_snake_game_loop)
    except Exception as e:
        print(f"\n\033[31mError in snake game: {e}\033[0m")
    finally:
        cleanup()


if __name__ == "__main__":
    print("\033[?25l\033[2J\033[H", end="")
    try:
        snake()
    finally:
        cleanup()



# installer.py

"""
TTYd binary and dependency installation.

This module handles the complete installation of ttyd and all its dependencies
across different platforms and environments. This updated version ensures terminaide
always uses its own internal ttyd binary without relying on system installations.

For Linux, it downloads pre-built binaries. For macOS, it compiles from source.
"""

import os
import stat
import logging
import platform
import subprocess
from pathlib import Path
from typing import Optional, Tuple, List
import urllib.request
import json

logger = logging.getLogger("terminaide")

# Fallback version in case API request fails
TTYD_FALLBACK_VERSION = "1.7.3"


def get_latest_ttyd_version() -> str:
    """
    Fetch the latest ttyd version from GitHub releases API.
    
    Returns:
        Latest version string (e.g., "1.7.7")
    """
    try:
        logger.info("Fetching latest ttyd version from GitHub API...")
        api_url = "https://api.github.com/repos/tsl0922/ttyd/releases/latest"
        
        with urllib.request.urlopen(api_url, timeout=10) as response:
            if response.status == 200:
                data = json.loads(response.read().decode())
                version = data.get("tag_name", "").lstrip("v")
                if version:
                    logger.info(f"Latest ttyd version: {version}")
                    return version
                else:
                    raise ValueError("No tag_name found in API response")
            else:
                raise urllib.error.HTTPError(api_url, response.status, "API request failed", None, None)
                
    except Exception as e:
        logger.warning(f"Failed to fetch latest ttyd version: {e}")
        logger.info(f"Using fallback version: {TTYD_FALLBACK_VERSION}")
        return TTYD_FALLBACK_VERSION


def get_ttyd_github_base(version: str) -> str:
    """Get the GitHub base URL for ttyd downloads for a specific version."""
    return f"https://github.com/tsl0922/ttyd/releases/download/{version}"

def get_platform_binaries(version: str) -> dict:
    """Get platform-specific binary URLs for a given ttyd version."""
    github_base = get_ttyd_github_base(version)
    return {
        ("Linux", "x86_64"): (f"{github_base}/ttyd.x86_64", "ttyd"),
        ("Linux", "aarch64"): (f"{github_base}/ttyd.aarch64", "ttyd"),
        ("Linux", "arm64"): (f"{github_base}/ttyd.aarch64", "ttyd"),
        # For macOS, we'll compile from source
    }

# Platform-specific system dependencies
SYSTEM_DEPENDENCIES = {
    "apt": {
        "packages": ["libwebsockets-dev", "libjson-c-dev"],
        "libraries": ["libwebsockets.so", "libjson-c.so"],
    },
    "brew": {
        "packages": ["libwebsockets", "json-c"],
        "libraries": ["libwebsockets.dylib", "libjson-c.dylib"],
    },
}


def get_package_manager() -> Optional[str]:
    """Detect the system's package manager."""
    if platform.system() == "Darwin":
        try:
            subprocess.check_output(["brew", "--version"])
            return "brew"
        except (subprocess.SubprocessError, FileNotFoundError):
            return None
    elif platform.system() == "Linux":
        try:
            subprocess.check_output(["apt-get", "--version"])
            return "apt"
        except (subprocess.SubprocessError, FileNotFoundError):
            return None
    return None


def install_system_dependencies(package_manager: str) -> None:
    """Install required system dependencies using the appropriate package manager."""
    deps = SYSTEM_DEPENDENCIES.get(package_manager)
    if not deps:
        raise RuntimeError(
            f"No dependency information for package manager: {package_manager}"
        )

    logger.info(f"Installing system dependencies using {package_manager}...")

    try:
        if package_manager == "apt":
            # Check if we can use sudo
            try:
                subprocess.check_output(["sudo", "-n", "true"])
                sudo_prefix = ["sudo"]
            except (subprocess.SubprocessError, FileNotFoundError):
                sudo_prefix = []

            # Update package list
            subprocess.run(
                [*sudo_prefix, "apt-get", "update", "-y"],
                check=True,
                capture_output=True,
            )

            # Install packages
            subprocess.run(
                [*sudo_prefix, "apt-get", "install", "-y", *deps["packages"]],
                check=True,
                capture_output=True,
            )

        elif package_manager == "brew":
            for pkg in deps["packages"]:
                subprocess.run(
                    ["brew", "install", pkg], check=True, capture_output=True
                )
    except subprocess.CalledProcessError as e:
        raise RuntimeError(
            f"Failed to install system dependencies using {package_manager}. "
            f"Error: {e.stderr.decode() if e.stderr else str(e)}"
        )


def verify_system_libraries(package_manager: str) -> List[str]:
    """Verify required system libraries are present and return any missing ones."""
    deps = SYSTEM_DEPENDENCIES.get(package_manager)
    if not deps:
        raise RuntimeError(
            f"No dependency information for package manager: {package_manager}"
        )

    missing = []
    if package_manager == "apt":
        try:
            output = subprocess.check_output(["ldconfig", "-p"]).decode()
            missing = [lib for lib in deps["libraries"] if lib not in output]
        except subprocess.SubprocessError:
            logger.warning("Could not verify libraries with ldconfig")
    elif package_manager == "brew":
        brew_prefix = subprocess.check_output(["brew", "--prefix"]).decode().strip()
        lib_path = Path(brew_prefix) / "lib"
        missing = [lib for lib in deps["libraries"] if not (lib_path / lib).exists()]

    return missing


def get_platform_info() -> Tuple[str, str]:
    """Get current platform and architecture."""
    system = platform.system()
    machine = platform.machine().lower()

    # Normalize ARM architecture names
    if machine in ["arm64", "aarch64"]:
        machine = "arm64"

    return system, machine


def get_binary_dir() -> Path:
    """
    Get the directory where the ttyd binary should be installed.

    Returns:
        Path to the binaries directory within the terminaide cache
    """
    package_dir = Path(__file__).parent.parent  # Go up from core to terminaide
    bin_dir = package_dir / "cache" / "binaries"
    bin_dir.mkdir(parents=True, exist_ok=True)
    return bin_dir


def download_binary(url: str, target_path: Path) -> None:
    """Download the ttyd binary from GitHub."""
    logger.info(f"Downloading ttyd from {url}")
    try:
        urllib.request.urlretrieve(url, target_path)
        # Make binary executable
        target_path.chmod(target_path.stat().st_mode | stat.S_IEXEC)
    except Exception as e:
        raise RuntimeError(f"Failed to download ttyd: {e}")


def compile_ttyd_from_source(target_path: Path, version: str) -> None:
    """
    Compile ttyd from source for macOS.

    This function downloads the ttyd source code, compiles it,
    and places the binary at the specified target path.
    """
    import tempfile
    import shutil
    import subprocess

    logger.info("Compiling ttyd from source for macOS...")

    # Store the original working directory
    original_dir = os.getcwd()

    try:
        # Create a temporary directory for compilation
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_dir_path = Path(temp_dir)
            source_tarball = temp_dir_path / "ttyd-source.tar.gz"

            # Download source code
            logger.info("Downloading ttyd source code...")
            source_url = f"https://github.com/tsl0922/ttyd/archive/refs/tags/{version}.tar.gz"
            logger.info(f"Source URL: {source_url}")
            try:
                urllib.request.urlretrieve(source_url, source_tarball)
            except Exception as e:
                raise RuntimeError(f"Failed to download ttyd source: {e}")

            # Extract source tarball
            logger.info("Extracting source code...")
            import tarfile

            with tarfile.open(source_tarball, "r:gz") as tar:
                tar.extractall(path=temp_dir_path)

            # The directory will be named "ttyd-{version}" when extracting from GitHub's tarball
            source_dir = temp_dir_path / f"ttyd-{version}"
            if not source_dir.exists():
                # Look for any ttyd-* directory as fallback
                for item in temp_dir_path.iterdir():
                    if item.is_dir() and item.name.startswith("ttyd"):
                        source_dir = item
                        break

            if not source_dir.exists():
                raise RuntimeError(
                    "Could not find the source directory after extraction"
                )

            # Create build directory
            build_dir = source_dir / "build"
            build_dir.mkdir(exist_ok=True)

            # Check for required dependencies
            logger.info("Checking for build dependencies...")
            required_cmds = ["cmake", "make", "cc"]
            missing_cmds = []

            for cmd in required_cmds:
                if not shutil.which(cmd):
                    missing_cmds.append(cmd)

            if missing_cmds:
                raise RuntimeError(
                    f"Missing required build tools: {', '.join(missing_cmds)}.\n"
                    f"Please install Xcode Command Line Tools with 'xcode-select --install'\n"
                    f"You may also need: brew install cmake libuv json-c libwebsockets openssl"
                )

            # Check for required libraries
            logger.info("Checking for required libraries...")
            try:
                # Check for libwebsockets, which is required for ttyd
                if not os.path.exists(
                    "/usr/local/opt/libwebsockets"
                ) and not os.path.exists("/opt/homebrew/opt/libwebsockets"):
                    logger.warning(
                        "libwebsockets not found, attempting to install with brew"
                    )
                    try:
                        subprocess.run(
                            ["brew", "install", "libwebsockets"],
                            check=True,
                            capture_output=True,
                        )
                    except subprocess.CalledProcessError:
                        logger.error("Failed to install libwebsockets with brew")
                        raise RuntimeError(
                            "Could not find libwebsockets, which is required for ttyd.\n"
                            "Please install it manually with: brew install libwebsockets"
                        )

                # Similarly check for json-c
                if not os.path.exists("/usr/local/opt/json-c") and not os.path.exists(
                    "/opt/homebrew/opt/json-c"
                ):
                    logger.warning("json-c not found, attempting to install with brew")
                    try:
                        subprocess.run(
                            ["brew", "install", "json-c"],
                            check=True,
                            capture_output=True,
                        )
                    except subprocess.CalledProcessError:
                        logger.error("Failed to install json-c with brew")
                        raise RuntimeError(
                            "Could not find json-c, which is required for ttyd.\n"
                            "Please install it manually with: brew install json-c"
                        )
            except Exception as e:
                logger.warning(f"Could not check or install dependencies: {e}")
                logger.warning("Will attempt to build anyway, but it might fail")

            # Configure and build
            logger.info("Configuring build...")
            try:
                # Change to build directory
                os.chdir(build_dir)

                # Run cmake with hints to find dependencies
                brew_prefix = None
                try:
                    brew_prefix = subprocess.check_output(
                        ["brew", "--prefix"], text=True
                    ).strip()
                    logger.info(f"Homebrew prefix: {brew_prefix}")
                except (subprocess.SubprocessError, FileNotFoundError):
                    logger.warning("Homebrew not found or error running brew --prefix")

                cmake_args = ["cmake", ".."]

                # Add hint paths for libraries if Homebrew is available
                if brew_prefix:
                    cmake_args.extend(
                        [
                            f"-DCMAKE_PREFIX_PATH={brew_prefix}",
                            f"-DOPENSSL_ROOT_DIR={brew_prefix}/opt/openssl",
                        ]
                    )

                logger.info(f"Running CMake with: {' '.join(cmake_args)}")
                subprocess.run(cmake_args, check=True, capture_output=True, text=True)

                # Run make
                logger.info("Building ttyd...")
                subprocess.run(["make"], check=True, capture_output=True, text=True)

                # Check if the binary was built successfully
                binary_path = build_dir / "ttyd"
                if not binary_path.exists():
                    raise RuntimeError("ttyd binary was not created after build")

                # Copy the binary to the target path
                shutil.copy2(binary_path, target_path)
                target_path.chmod(target_path.stat().st_mode | stat.S_IEXEC)

                logger.info(
                    f"Successfully compiled ttyd and installed to {target_path}"
                )

            except subprocess.CalledProcessError as e:
                error_output = e.stderr
                raise RuntimeError(f"Build failed: {error_output}")
            except Exception as e:
                raise RuntimeError(f"Error during build process: {e}")
    finally:
        # Always restore the original working directory
        os.chdir(original_dir)


def get_ttyd_path(force_reinstall: bool = False) -> Optional[Path]:
    """
    Get path to installed ttyd binary, installing if necessary.

    Args:
        force_reinstall: If True, reinstall ttyd even if it exists

    Returns:
        Path to the ttyd binary
    """
    # Get the latest version
    version = get_latest_ttyd_version()
    
    system, machine = get_platform_info()
    platform_key = (system, machine)
    binary_dir = get_binary_dir()
    binary_name = "ttyd"
    binary_path = binary_dir / binary_name

    # Handle macOS specially - we'll compile from source
    if system == "Darwin":
        logger.info("macOS detected - ttyd will be compiled from source")

        # Check if binary exists and is executable, skip if not force_reinstall
        if (
            not force_reinstall
            and binary_path.exists()
            and os.access(binary_path, os.X_OK)
        ):
            logger.info(f"Using existing ttyd binary at: {binary_path}")
            return binary_path

        # Compile from source for macOS
        try:
            compile_ttyd_from_source(binary_path, version)
            return binary_path
        except Exception as e:
            logger.error(f"Error compiling ttyd from source: {e}")
            raise RuntimeError(f"Failed to compile ttyd for macOS: {e}")

    # For other platforms, continue with binary download approach
    platform_binaries = get_platform_binaries(version)
    
    # Try common platform keys for Linux
    if platform_key not in platform_binaries:
        if system == "Linux":
            for machine_type in ["arm64", "aarch64", "x86_64"]:
                alt_key = (system, machine_type)
                if alt_key in platform_binaries:
                    platform_key = alt_key
                    break

    if platform_key not in platform_binaries:
        raise PlatformNotSupportedError(system, machine)

    # Set up system dependencies first
    package_manager = get_package_manager()
    if not package_manager:
        raise RuntimeError(
            f"No supported package manager found for {system}. "
            "Please install libwebsockets and json-c manually."
        )

    # Check for missing libraries
    missing_libs = verify_system_libraries(package_manager)
    if missing_libs:
        install_system_dependencies(package_manager)
        # Verify installation succeeded
        still_missing = verify_system_libraries(package_manager)
        if still_missing:
            raise RuntimeError(
                f"Failed to install required libraries: {', '.join(still_missing)}"
            )

    url, download_binary_name = platform_binaries[platform_key]

    # Check if binary exists and is executable, or if force_reinstall is specified
    if (
        force_reinstall
        or not binary_path.exists()
        or not os.access(binary_path, os.X_OK)
    ):
        download_binary(url, binary_path)

    return binary_path


def setup_ttyd(force_reinstall: bool = None) -> Path:
    """
    Ensure ttyd is installed and return its path.

    This is the main entry point for the installer module.

    Args:
        force_reinstall: If True, reinstall ttyd even if it exists.
                         If None, check TERMINAIDE_FORCE_REINSTALL env var.

    Returns:
        Path to the ttyd binary
    """
    # Check environment variable if force_reinstall not specified
    if force_reinstall is None:
        force_reinstall = os.environ.get("TERMINAIDE_FORCE_REINSTALL", "").lower() in (
            "1",
            "true",
            "yes",
        )

    try:
        # Get our managed version
        binary_path = get_ttyd_path(force_reinstall)
        if binary_path and os.access(binary_path, os.X_OK):
            return binary_path

        raise RuntimeError("Failed to locate or install ttyd")

    except Exception as e:
        logger.error(f"Failed to set up ttyd: {e}")
        raise


class PlatformNotSupportedError(RuntimeError):
    """Raised when trying to install on an unsupported platform."""

    def __init__(self, system: str, machine: str):
        message = (
            f"Platform not supported: {system} {machine}\n"
            f"Terminaide currently supports: Linux x86_64, Linux ARM64, macOS.\n"
            f"Please file an issue if you need support for other platforms."
        )
        super().__init__(message)
        self.system = system
        self.machine = machine



# tetris.py

import curses
import random
import signal
import sys
import time
import os
from collections import deque

stdscr = None
exit_requested = False

TETROMINOS = [
    [[(0, 0), (0, 1), (0, 2), (0, 3)], [(0, 0), (1, 0), (2, 0), (3, 0)]],
    [[(0, 0), (0, 1), (1, 0), (1, 1)]],
    [
        [(0, 1), (1, 0), (1, 1), (1, 2)],
        [(0, 0), (1, 0), (1, 1), (2, 0)],
        [(0, 0), (0, 1), (0, 2), (1, 1)],
        [(0, 1), (1, 0), (1, 1), (2, 1)],
    ],
    [
        [(0, 0), (1, 0), (2, 0), (2, 1)],
        [(0, 0), (0, 1), (0, 2), (1, 0)],
        [(0, 0), (0, 1), (1, 1), (2, 1)],
        [(0, 2), (1, 0), (1, 1), (1, 2)],
    ],
    [
        [(0, 1), (1, 1), (2, 0), (2, 1)],
        [(0, 0), (1, 0), (1, 1), (1, 2)],
        [(0, 0), (0, 1), (1, 0), (2, 0)],
        [(0, 0), (0, 1), (0, 2), (1, 2)],
    ],
    [[(0, 1), (0, 2), (1, 0), (1, 1)], [(0, 0), (1, 0), (1, 1), (2, 1)]],
    [[(0, 0), (0, 1), (1, 1), (1, 2)], [(0, 1), (1, 0), (1, 1), (2, 0)]],
]
TETROMINO_COLORS = [
    curses.COLOR_CYAN,
    curses.COLOR_YELLOW,
    curses.COLOR_MAGENTA,
    curses.COLOR_WHITE,
    curses.COLOR_BLUE,
    curses.COLOR_GREEN,
    curses.COLOR_RED,
]


def _tetris_game_loop(stdscr_param):
    """Main tetris game function that handles the game loop.

    Args:
        stdscr_param: The curses window.
    """
    global stdscr, exit_requested
    stdscr = stdscr_param
    exit_requested = False

    signal.signal(signal.SIGINT, handle_exit)
    stdscr.clear()
    stdscr.refresh()
    os.system("clear" if os.name == "posix" else "cls")
    setup_terminal(stdscr)
    max_y, max_x = stdscr.getmaxyx()
    high_score = 0

    while True:
        if exit_requested:
            cleanup()
            return

        score = run_game(stdscr, max_y, max_x, high_score)

        if exit_requested:
            cleanup()
            return

        high_score = max(high_score, score)
        if show_game_over(stdscr, score, high_score, max_y, max_x):
            break


def setup_terminal(stdscr):
    """Configure terminal settings for the game."""
    curses.curs_set(0)
    curses.noecho()
    curses.cbreak()
    stdscr.keypad(True)
    curses.use_env(False)
    curses.start_color()
    curses.use_default_colors()
    for i, c in enumerate(TETROMINO_COLORS):
        curses.init_pair(i + 1, c, -1)
    curses.init_pair(8, curses.COLOR_WHITE, -1)
    curses.init_pair(9, curses.COLOR_YELLOW, -1)


def run_game(stdscr, max_y, max_x, high_score=0):
    """Run the main tetris game loop.

    Args:
        stdscr: The curses window.
        max_y, max_x: Maximum y and x coordinates.
        high_score: Current high score.

    Returns:
        int: The final score.
    """
    global exit_requested

    board_height = min(20, max_y - 7)
    board_width = min(10, max_x // 2 - 2)
    start_y = 2
    start_x = (max_x - board_width * 2) // 2
    board = [[0 for _ in range(board_width)] for _ in range(board_height)]
    score = 0
    level = 1
    lines_cleared = 0
    game_speed = 500
    game_win = curses.newwin(board_height + 2, board_width * 2 + 2, start_y, start_x)
    game_win.keypad(True)
    game_win.timeout(game_speed)
    piece_type = random.randint(0, len(TETROMINOS) - 1)
    rotation = 0
    pos = [0, board_width // 2 - 1]
    next_piece = random.randint(0, len(TETROMINOS) - 1)
    next_win = curses.newwin(6, 10, start_y, board_width * 2 + start_x + 4)
    fall_time = 0
    last_move = time.time()

    while True:
        if exit_requested:
            return score

        now = time.time()
        delta = now - last_move
        last_move = now
        fall_time += delta * 1000

        key = game_win.getch()

        # Check for exit key
        if key in (ord("q"), 27):  # q or ESC
            return score

        if key in [curses.KEY_LEFT, ord("a"), ord("A")]:
            np = [pos[0], pos[1] - 1]
            if is_valid_position(board, TETROMINOS[piece_type][rotation], np):
                pos = np
        elif key in [curses.KEY_RIGHT, ord("d"), ord("D")]:
            np = [pos[0], pos[1] + 1]
            if is_valid_position(board, TETROMINOS[piece_type][rotation], np):
                pos = np
        elif key in [curses.KEY_DOWN, ord("s"), ord("S")]:
            np = [pos[0] + 1, pos[1]]
            if is_valid_position(board, TETROMINOS[piece_type][rotation], np):
                pos = np
                score += 1
                fall_time = 0
        elif key in [curses.KEY_UP, ord("w"), ord("W")]:
            r = (rotation + 1) % len(TETROMINOS[piece_type])
            if is_valid_position(board, TETROMINOS[piece_type][r], pos):
                rotation = r
        elif key == ord(" "):
            while is_valid_position(
                board, TETROMINOS[piece_type][rotation], [pos[0] + 1, pos[1]]
            ):
                pos[0] += 1
                score += 2
            fall_time = game_speed + 1

        if fall_time >= game_speed:
            np = [pos[0] + 1, pos[1]]
            if is_valid_position(board, TETROMINOS[piece_type][rotation], np):
                pos = np
            else:
                place_tetromino(
                    board, TETROMINOS[piece_type][rotation], pos, piece_type + 1
                )
                c = clear_lines(board)
                if c > 0:
                    lines_cleared += c
                    score += calculate_score(c, level)
                    level = lines_cleared // 10 + 1
                    game_speed = max(100, 500 - (level - 1) * 50)
                    game_win.timeout(game_speed)
                piece_type = next_piece
                rotation = 0
                pos = [0, board_width // 2 - 1]
                next_piece = random.randint(0, len(TETROMINOS) - 1)
                if not is_valid_position(board, TETROMINOS[piece_type][rotation], pos):
                    return score
            fall_time = 0

        draw_game(
            stdscr,
            game_win,
            next_win,
            board,
            TETROMINOS[piece_type][rotation],
            pos,
            piece_type,
            next_piece,
            score,
            level,
            lines_cleared,
            high_score,
            board_height,
            board_width,
            max_x,
            start_y,
            start_x,
        )


def is_valid_position(board, piece, pos):
    """Check if the tetromino position is valid.

    Args:
        board: Game board state.
        piece: Tetromino piece shape.
        pos: Position [y, x] to check.

    Returns:
        bool: True if position is valid, False otherwise.
    """
    bh = len(board)
    bw = len(board[0])
    for y, x in piece:
        ny = pos[0] + y
        nx = pos[1] + x
        if nx < 0 or nx >= bw or ny >= bh:
            return False
        if ny >= 0 and board[ny][nx] != 0:
            return False
    return True


def place_tetromino(board, piece, pos, t):
    """Place a tetromino on the board.

    Args:
        board: Game board state.
        piece: Tetromino piece shape.
        pos: Position [y, x] to place.
        t: Tetromino type (for color).
    """
    for y, x in piece:
        ny = pos[0] + y
        nx = pos[1] + x
        if ny >= 0:
            board[ny][nx] = t


def clear_lines(board):
    """Clear completed lines and calculate score.

    Args:
        board: Game board state.

    Returns:
        int: Number of lines cleared.
    """
    bh = len(board)
    bw = len(board[0])
    c = 0
    y = bh - 1
    while y >= 0:
        if all(board[y][x] != 0 for x in range(bw)):
            for y2 in range(y, 0, -1):
                for x in range(bw):
                    board[y2][x] = board[y2 - 1][x]
            for x in range(bw):
                board[0][x] = 0
            c += 1
        else:
            y -= 1
    return c


def calculate_score(lines, level):
    """Calculate score based on lines cleared and level.

    Args:
        lines: Number of lines cleared at once.
        level: Current level.

    Returns:
        int: Score to add.
    """
    s = [0, 100, 300, 500, 800]
    return s[min(lines, 4)] * level


def safe_addstr(stdscr, y, x, text, attr=0):
    """Safely add a string to the screen, handling boundary conditions.

    Args:
        stdscr: The screen to draw on.
        y, x: Coordinates to start drawing.
        text: Text to draw.
        attr: Text attributes.
    """
    h, w = stdscr.getmaxyx()
    if y < 0 or y >= h or x < 0 or x >= w:
        return
    ml = w - x
    if ml <= 0:
        return
    t = text[:ml]
    try:
        stdscr.addstr(y, x, t, attr)
    except:
        curses.error


def draw_game(
    stdscr,
    gw,
    nw,
    board,
    piece,
    pos,
    ptype,
    ntype,
    score,
    level,
    lines,
    high_score,
    bh,
    bw,
    mx,
    sy,
    sx,
):
    """Draw the game screen with all elements.

    Args:
        stdscr: Main screen.
        gw: Game window.
        nw: Next piece window.
        board: Game board state.
        piece: Current tetromino shape.
        pos: Current position [y, x].
        ptype: Current piece type.
        ntype: Next piece type.
        score: Current score.
        level: Current level.
        lines: Lines cleared.
        high_score: High score.
        bh, bw: Board height and width.
        mx: Max x coordinate.
        sy, sx: Start y and x coordinates.
    """
    gw.erase()
    nw.erase()
    gw.box()
    nw.box()
    t = "PLAY TETRIS"
    ww = gw.getmaxyx()[1]
    if ww > len(t) + 4:
        gw.addstr(0, (ww - len(t)) // 2, t, curses.A_BOLD | curses.color_pair(9))
    nw.addstr(0, 1, "NEXT", curses.A_BOLD | curses.color_pair(9))
    for y in range(bh):
        for x in range(bw):
            c = board[y][x]
            if c != 0:
                try:
                    gw.addstr(
                        y + 1, x * 2 + 1, "[]", curses.color_pair(c) | curses.A_BOLD
                    )
                except:
                    curses.error
    for y, x in piece:
        ny = pos[0] + y
        nx = pos[1] + x
        if ny >= 0:
            try:
                gw.addstr(
                    ny + 1,
                    nx * 2 + 1,
                    "[]",
                    curses.color_pair(ptype + 1) | curses.A_BOLD,
                )
            except:
                curses.error
    nt = TETROMINOS[ntype][0]
    minx = min(a for _, a in nt)
    maxx = max(a for _, a in nt)
    miny = min(a for a, _ in nt)
    maxy = max(a for a, _ in nt)
    cy = 3
    cx = 5
    for yy, xx in nt:
        dy = cy + yy - (miny + maxy) // 2
        dx = cx + (xx - (minx + maxx) // 2) * 2
        try:
            nw.addstr(dy, dx, "[]", curses.color_pair(ntype + 1) | curses.A_BOLD)
        except:
            curses.error
    safe_addstr(stdscr, 0, 0, " " * mx)
    st = f" Score: {score} "
    lt = f" Level: {level} "
    hi = f" High: {high_score} "
    safe_addstr(stdscr, 0, 2, st, curses.color_pair(9) | curses.A_BOLD)
    safe_addstr(stdscr, 0, 2 + len(st) + 2, lt, curses.color_pair(9) | curses.A_BOLD)
    sw = stdscr.getmaxyx()[1]
    hx = min(sw - len(hi) - 1, mx - len(hi) - 2)
    safe_addstr(stdscr, 0, hx, hi, curses.color_pair(9) | curses.A_BOLD)
    # Controls below the board window
    ctrl = "↑:Rotate  ←→:Move  ↓:Drop  Space:Hard Drop  Q:Quit"
    controls_row = sy + bh + 3  # Just a few lines below bottom of game board
    controls_col = sx + ((bw * 2 + 2) - len(ctrl)) // 2
    safe_addstr(stdscr, controls_row, controls_col, ctrl)

    stdscr.noutrefresh()
    gw.noutrefresh()
    nw.noutrefresh()
    curses.doupdate()


def show_game_over(stdscr, score, high_score, my, mx):
    """Show game over screen and handle restart/quit options.

    Args:
        stdscr: The screen to draw on.
        score: Final score.
        high_score: High score.
        my, mx: Maximum y and x coordinates.

    Returns:
        bool: True if user chooses to quit, False to restart.
    """
    stdscr.clear()
    cy = my // 2
    data = [
        ("GAME OVER", -3, curses.A_BOLD | curses.color_pair(3)),
        (f"Your Score: {score}", -1, curses.color_pair(9)),
        (f"High Score: {high_score}", 0, curses.color_pair(9)),
        ("Press 'r' to restart", 2, 0),
        ("Press 'q' to quit", 3, 0),
    ]
    for txt, yo, attr in data:
        stdscr.addstr(cy + yo, mx // 2 - len(txt) // 2, txt, attr)
    stdscr.noutrefresh()
    curses.doupdate()
    stdscr.nodelay(False)
    while True:
        k = stdscr.getch()
        if k == ord("q"):
            return True
        if k == ord("r"):
            return False


def cleanup():
    """Clean up terminal state when exiting."""
    if stdscr:
        try:
            curses.endwin()
            print("\033[?25l\033[2J\033[H", end="")
            try:
                rows, cols = stdscr.getmaxyx()
            except:
                rows, cols = 24, 80
            msg = "Thanks for playing Tetris!"
            print("\033[2;{}H{}".format((cols - len(msg)) // 2, msg))
            print("\033[3;{}H{}".format((cols - len("Goodbye!")) // 2, "Goodbye!"))
            sys.stdout.flush()
        except:
            pass


def handle_exit(sig, frame):
    """Handle SIGINT (Ctrl+C) for program exit."""
    global exit_requested
    exit_requested = True


def tetris():
    """Run the tetris game.

    This is the main public-facing function for launching the tetris game.
    """
    try:
        os.system("clear" if os.name == "posix" else "cls")
        print("\033[2J\033[H", end="")
        sys.stdout.flush()
        curses.wrapper(_tetris_game_loop)
    except Exception as e:
        print(f"\n\033[31mError in tetris game: {e}\033[0m")
    finally:
        cleanup()


if __name__ == "__main__":
    print("\033[?25l\033[2J\033[H", end="")
    sys.stdout.flush()
    os.system("clear" if os.name == "posix" else "cls")
    try:
        tetris()
    finally:
        cleanup()



# logger.py

"""Logging configuration and terminal output formatting for Terminaide."""

import sys
import logging
import hashlib
from typing import Dict, Optional
from pathlib import Path


# Enhanced ANSI color codes for logging levels
LEVEL_COLORS = {
    "DEBUG": "\033[36m",  # Cyan
    "INFO": "\033[32m",  # Green
    "WARNING": "\033[33m",  # Yellow
    "ERROR": "\033[31m",  # Red
    "CRITICAL": "\033[41m",  # Red background
    "RESET": "\033[0m",  # Reset colors
}


class ColorAlignedFormatter(logging.Formatter):
    """Logging formatter that adds colors and aligns log levels."""

    def format(self, record):
        levelname = record.levelname
        # Ensure the level name + padding + colon takes exactly 10 characters
        padding_length = max(1, 9 - len(levelname))
        padding = " " * padding_length

        # Add colors if the system supports it
        if sys.stdout.isatty():  # Only apply colors if running in a terminal
            colored_levelname = (
                f"{LEVEL_COLORS.get(levelname, '')}{levelname}{LEVEL_COLORS['RESET']}"
            )
            return f"{colored_levelname}:{padding}{record.getMessage()}"
        else:
            return f"{levelname}:{padding}{record.getMessage()}"


class RouteColorManager:
    """Manages consistent color assignments and formatting for routes."""

    # Bright, distinguishable colors for route titles
    ROUTE_COLORS = [
        "\033[94m",  # Bright Blue
        "\033[95m",  # Bright Magenta
        "\033[96m",  # Bright Cyan
        "\033[93m",  # Bright Yellow
        "\033[91m",  # Bright Red
        "\033[92m",  # Bright Green
        "\033[97m",  # Bright White
        "\033[35m",  # Magenta
        "\033[34m",  # Blue
        "\033[36m",  # Cyan
        "\033[33m",  # Yellow
        "\033[31m",  # Red
        "\033[32m",  # Green
        "\033[37m",  # White
        "\033[90m",  # Bright Black (Gray)
        "\033[38;5;208m",  # Orange
        "\033[38;5;165m",  # Pink
        "\033[38;5;51m",  # Light Cyan
        "\033[38;5;226m",  # Light Yellow
        "\033[38;5;46m",  # Light Green
    ]

    RESET = "\033[0m"

    def __init__(self):
        self._route_colors: Dict[str, str] = {}
        self._color_enabled = sys.stdout.isatty()

    def get_route_color(self, route_path: str) -> str:
        """Get a consistent color for a route based on its path."""
        if not self._color_enabled:
            return ""

        if route_path not in self._route_colors:
            # Use hash to get consistent color assignment
            hash_value = int(hashlib.md5(route_path.encode()).hexdigest()[:8], 16)
            color_index = hash_value % len(self.ROUTE_COLORS)
            self._route_colors[route_path] = self.ROUTE_COLORS[color_index]

        return self._route_colors[route_path]

    def colorize_title(self, title: str, route_path: str) -> str:
        """Colorize a route title with its assigned color."""
        if not self._color_enabled:
            return title
        color = self.get_route_color(route_path)
        return f"{color}{title}{self.RESET}"

    def format_route_info(
        self,
        route_path: str,
        title: str,
        script_config,
        port: Optional[int] = None,
        pid: Optional[int] = None,
    ) -> tuple[str, str]:
        """Create a standardized route info string with consistent formatting.

        Returns:
            tuple: (main_line, script_line) for separate logging
        """
        colored_title = self.colorize_title(title, route_path)
        route_type = "function" if script_config.is_function_based else "script"

        # Format the main info line in the new style
        main_line = f"Serving '{colored_title}' {route_type}"

        if pid:
            # Colorize PID in cyan like uvicorn
            if self._color_enabled:
                colored_pid = f"\033[36m{pid}\033[0m"  # Cyan color
                main_line += f" [{colored_pid}]"
            else:
                main_line += f" [{pid}]"

        if port:
            main_line += f" on port {port}"

        main_line += f" at '{route_path}'"

        # Format the script path line
        script_info = (
            str(script_config.effective_script_path)
            if script_config.effective_script_path
            else "function"
        )
        script_line = f"Script path: {script_info}"

        return main_line, script_line


# Global instance
route_color_manager = RouteColorManager()


# Legacy compatibility functions (used by __init__.py)
def get_route_color(route_path: str) -> str:
    """Get a consistent color for a route based on its path."""
    return route_color_manager.get_route_color(route_path)


def colorize_route_title(title: str, route_path: str) -> str:
    """Colorize a route title with its assigned color."""
    return route_color_manager.colorize_title(title, route_path)


def setup_package_logging(log_level="info"):
    """Configure package-level logging for Terminaide.

    Args:
        log_level: Logging level as string ("debug", "info", "warning", "error", "critical")
                  or None to skip configuration. Defaults to "info".
    """
    logger = logging.getLogger("terminaide")

    if log_level is not None:
        # Only add handler if none exist (avoid duplicate handlers)
        if not logger.handlers:
            handler = logging.StreamHandler()
            handler.setFormatter(ColorAlignedFormatter())
            logger.addHandler(handler)

            # Convert string level to logging constant
            level_map = {
                "debug": logging.DEBUG,
                "info": logging.INFO,
                "warning": logging.WARNING,
                "error": logging.ERROR,
                "critical": logging.CRITICAL,
            }
            logger.setLevel(level_map.get(log_level.lower(), logging.INFO))
            logger.propagate = False  # Prevent propagation to root logger

    return logger



# exceptions.py

"""
Custom exceptions for the terminaide package.

These exceptions provide specific error cases that may occur during
ttyd setup, installation, and operation. The exception hierarchy has been
expanded to support multi-terminal (apps-server) routing scenarios.
"""

from typing import Optional, List
from pathlib import Path


class terminaideError(Exception):
    """Base exception for all terminaide errors."""


class BinaryError(terminaideError):
    """Base class for binary-related errors."""

    def __init__(self, message: str, binary_path: Optional[Path] = None):
        super().__init__(message)
        self.binary_path = binary_path


class InstallationError(BinaryError):
    """Raised when ttyd binary installation fails."""

    def __init__(
        self,
        message: str,
        binary_path: Optional[Path] = None,
        platform: Optional[str] = None,
    ):
        super().__init__(
            f"Installation failed: {message}"
            + (f" (platform: {platform})" if platform else ""),
            binary_path,
        )
        self.platform = platform


class PlatformNotSupportedError(InstallationError):
    """Raised when trying to install on an unsupported platform."""

    def __init__(self, system: str, machine: str):
        super().__init__(
            f"Platform not supported: {system} {machine}",
            platform=f"{system} {machine}",
        )
        self.system = system
        self.machine = machine


class DependencyError(InstallationError):
    """Raised when required system dependencies are missing."""

    def __init__(self, missing_deps: list[str]):
        deps_str = ", ".join(missing_deps)
        super().__init__(
            f"Missing required dependencies: {deps_str}\n"
            "Please install:\n"
            "  Ubuntu/Debian: apt-get install libwebsockets-dev libjson-c-dev\n"
            "  MacOS: brew install libwebsockets json-c"
        )
        self.missing_deps = missing_deps


class DownloadError(InstallationError):
    """Raised when downloading the ttyd binary fails."""

    def __init__(self, url: str, error: str):
        super().__init__(f"Failed to download from {url}: {error}")
        self.url = url
        self.error = error


class TTYDStartupError(BinaryError):
    """Raised when ttyd process fails to start."""

    def __init__(
        self,
        message: str = None,
        stderr: str = None,
        binary_path: Optional[Path] = None,
        route_path: Optional[str] = None,
    ):
        route_info = f" for route '{route_path}'" if route_path else ""
        msg = message or f"Failed to start ttyd process{route_info}"
        if stderr:
            msg = f"{msg}\nttyd error output:\n{stderr}"
        super().__init__(msg, binary_path)
        self.stderr = stderr
        self.route_path = route_path


class TTYDProcessError(BinaryError):
    """Raised when ttyd process encounters an error during operation."""

    def __init__(
        self,
        message: str = None,
        exit_code: int = None,
        binary_path: Optional[Path] = None,
        route_path: Optional[str] = None,
    ):
        route_info = f" for route '{route_path}'" if route_path else ""
        msg = message or f"ttyd process error{route_info}"
        if exit_code is not None:
            msg = f"{msg} (exit code: {exit_code})"
        super().__init__(msg, binary_path)
        self.exit_code = exit_code
        self.route_path = route_path


class ClientScriptError(terminaideError):
    """Raised when there are issues with the client script."""

    def __init__(
        self, script_path: str, message: str = None, route_path: Optional[str] = None
    ):
        route_info = f" for route '{route_path}'" if route_path else ""
        super().__init__(
            f"Error with client script '{script_path}'{route_info}: {message or 'Unknown error'}"
        )
        self.script_path = script_path
        self.route_path = route_path


class TemplateError(terminaideError):
    """Raised when there are issues with the HTML template."""

    def __init__(self, template_path: str = None, message: str = None):
        msg = "Template error"
        if template_path:
            msg = f"{msg} with '{template_path}'"
        if message:
            msg = f"{msg}: {message}"
        super().__init__(msg)
        self.template_path = template_path


class ProxyError(terminaideError):
    """Raised when there are issues with the proxy configuration or operation."""

    def __init__(
        self,
        message: str = None,
        original_error: Exception = None,
        route_path: Optional[str] = None,
    ):
        route_info = f" for route '{route_path}'" if route_path else ""
        msg = (message or "Proxy error") + route_info
        if original_error:
            msg = f"{msg}: {str(original_error)}"
        super().__init__(msg)
        self.original_error = original_error
        self.route_path = route_path


class ConfigurationError(terminaideError):
    """Raised when there are issues with the provided configuration."""

    def __init__(self, message: str, field: str = None):
        msg = f"Configuration error"
        if field:
            msg = f"{msg} in '{field}'"
        msg = f"{msg}: {message}"
        super().__init__(msg)
        self.field = field


# --- New Exceptions for Multi-Script Support ---


class RouteNotFoundError(ProxyError):
    """Raised when a request cannot be routed to any known script configuration."""

    def __init__(self, message: str, request_path: Optional[str] = None):
        path_info = f" (request path: {request_path})" if request_path else ""
        super().__init__(f"Route not found: {message}{path_info}")
        self.request_path = request_path


class PortAllocationError(ConfigurationError):
    """Raised when port allocation fails for multi-terminal (apps-server) configuration."""

    def __init__(self, message: str, attempted_ports: Optional[List[int]] = None):
        port_info = f" (attempted ports: {attempted_ports})" if attempted_ports else ""
        super().__init__(f"{message}{port_info}", field="port")
        self.attempted_ports = attempted_ports


class ScriptConfigurationError(ConfigurationError):
    """Raised when there are issues with script configurations."""

    def __init__(self, message: str, route_path: Optional[str] = None):
        field = f"terminal_routes[{route_path}]" if route_path else "terminal_routes"
        super().__init__(message, field=field)
        self.route_path = route_path


class DuplicateRouteError(ScriptConfigurationError):
    """Raised when duplicate route paths are found in script configurations."""

    def __init__(self, route_path: str):
        super().__init__(f"Duplicate route path: {route_path}", route_path=route_path)
